# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform draft endpoint operations."""
from __future__ import annotations

import datetime
from typing import Any, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import (
    BackendDatabaseError,
    DraftAlreadyExistsError,
    DraftDoesNotExistError,
    EntityDoesNotExistError,
)
from dioptra.restapi.v1.groups.service import GroupIdService

LOGGER: BoundLogger = structlog.stdlib.get_logger()


class ResourceDraftsService(object):
    """The service methods for managing resource drafts."""

    @inject
    def __init__(
        self,
        resource_type: str,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the draft service.

        All arguments are provided via dependency injection.

        Args:
            group_id_service: A GroupIdService object.
        """
        self._resource_type = resource_type
        self._group_id_service = group_id_service

    def get(
        self,
        draft_type: str,
        group_id: int | None,
        base_resource_id: int | None,
        page_index: int,
        page_length: int,
        **kwargs,
    ) -> Any:
        """Fetch a list of drafts

        Args:
            draft_type: The type of drafts to retrieve (all, existing, or new)
            group_id: A group ID used to filter results.
            base_resource_id: An additional resource ID used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of drafts to be returned.

        Returns:
            A tuple containing a list of drafts and the total number of drafts.

        Raises:
            BackendDatabaseError: If the database query returns a None when counting
                the number of drafts.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of drafts")

        filters = list()

        if group_id is not None:
            filters.append(models.DraftResource.group_id == group_id)

        if base_resource_id is not None:
            filters.append(
                models.DraftResource.payload["base_resource_id"]
                .as_string()
                .cast(Integer)
                == base_resource_id
            )

        if draft_type == "existing":
            filters.append(
                models.DraftResource.payload["resource_id"].as_string()
                != None  # noqa: E711
            )
        elif draft_type == "new":
            filters.append(
                models.DraftResource.payload["resource_id"].as_string()
                == None  # noqa: E711
            )

        stmt = select(func.count(models.DraftResource.draft_resource_id)).where(
            *filters,
            models.DraftResource.resource_type == self._resource_type,
            models.DraftResource.user_id == current_user.user_id,
        )
        total_num_drafts = db.session.scalars(stmt).first()

        if total_num_drafts is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_drafts == 0:
            return [], total_num_drafts

        stmt = (
            select(models.DraftResource)
            .filter(  # type: ignore
                *filters,
                models.DraftResource.resource_type == self._resource_type,
                models.DraftResource.user_id == current_user.user_id,
            )
            .offset(page_index)
            .limit(page_length)
        )
        drafts = db.session.scalars(stmt).all()

        return drafts, total_num_drafts

    def create(
        self,
        base_resource_id: int | None,
        payload: dict[str, Any],
        commit: bool = True,
        **kwargs,
    ) -> models.DraftResource:
        """Create a new draft.

        Args:
            payload: The contents of the draft.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created draft object.

        Raises:
            EntityDoesNotExistError: If the group with the provided ID does not exist.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        group_id = payload.pop("group_id", None)

        if base_resource_id is None:
            group = self._group_id_service.get(group_id, error_if_not_found=True)
        else:
            stmt = select(models.Resource).filter_by(
                resource_id=base_resource_id, is_deleted=False
            )
            resource = db.session.scalar(stmt)
            if resource is None:
                raise EntityDoesNotExistError(
                    self._resource_type, resource_id=base_resource_id
                )
            group = resource.owner

        draft_payload = {
            "resource_data": payload,
            "resource_id": None,
            "resource_snapshot_id": None,
            "base_resource_id": base_resource_id,
        }

        draft = models.DraftResource(
            resource_type=self._resource_type,
            target_owner=group,
            payload=draft_payload,
            creator=current_user,
        )
        db.session.add(draft)

        if commit:
            db.session.commit()
            log.debug("Draft creation successful", resource_id=draft.draft_resource_id)

        return draft


class ResourceDraftsIdService(object):
    """The service methods for managing a specific resource draft."""

    @inject
    def __init__(
        self,
        resource_type: str,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the draft service.

        All arguments are provided via dependency injection.

        Args:
            group_id_service: A GroupIdService object.
        """
        self._resource_type = resource_type
        self._group_id_service = group_id_service

    def get(
        self,
        draft_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.DraftResource | None:
        """Fetch a draft by its unique id.

        Args:
            resource_id: The unique id of the resource.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.

        Returns:
            The draft object if found, otherwise None.

        Raises:
            DraftDoesNotExistError: If the draft is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get draft by id", draft_id=draft_id)

        stmt = select(models.DraftResource).filter(
            models.DraftResource.resource_type == self._resource_type,
            models.DraftResource.draft_resource_id == draft_id,
            models.DraftResource.user_id == current_user.user_id,
        )
        draft = db.session.scalars(stmt).first()

        if draft is None:
            if error_if_not_found:
                raise DraftDoesNotExistError(draft_resource_id=draft_id)

            return None

        return draft

    def modify(
        self,
        draft_id: int,
        payload: dict[str, Any],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> models.DraftResource | None:
        """Modify a Draft

        Args:
            resource_id: The unique id of the resource.
            payload: The contents of the draft.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated draft resource object.

        Raises:
            DraftDoesNotExistError: If the draft is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        draft = self.get(draft_id, error_if_not_found=error_if_not_found, log=log)

        if draft is None:
            return None

        current_timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        draft.payload["resource_data"] = payload
        draft.last_modified_on = current_timestamp

        if commit:
            db.session.commit()
            log.debug("Draft modification successful", draft_resource_id=draft_id)

        return draft

    def delete(self, draft_id: int, **kwargs) -> dict[str, Any]:
        """Delete a draft.

        Args:
            draft_id: The unique id of the draft.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        draft = self.get(draft_id, error_if_not_found=True, log=log)
        db.session.delete(draft)
        db.session.commit()
        log.debug("Draft deleted", draft_resource_id=draft_id)

        return {"status": "Success", "id": [draft_id]}


class ResourceIdDraftService(object):
    """The service methods for managing the draft for an existing resource."""

    def __init__(self, resource_type: str):
        self._resource_type = resource_type

    def get(
        self,
        resource_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> tuple[models.DraftResource | None, int]:
        """Fetch a draft by its unique id.

        Args:
            resource_id: The unique id of the draft.
            error_if_not_found: If True, raise an error if the draft is not found.
                Defaults to False.

        Returns:
            The draft object if found, otherwise None.

        Raises:
            DraftDoesNotExistError: If the draft is not found and `error_if_not_found`
                is True.
            BackendDatabaseError: If the database query returns a None when counting
                the number of drafts.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get draft by resource id", resource_id=resource_id)

        stmt = select(func.count(models.DraftResource.draft_resource_id)).where(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
            == resource_id,
            models.DraftResource.user_id != current_user.user_id,
        )
        num_other_drafts = db.session.scalars(stmt).first()

        if num_other_drafts is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        stmt = select(models.DraftResource).filter(  # type: ignore
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
            == resource_id,
            models.DraftResource.user_id == current_user.user_id,
        )
        draft = cast(models.DraftResource, db.session.scalars(stmt).first())

        if draft is None:
            if error_if_not_found:
                raise DraftDoesNotExistError(resource_id=resource_id)

            return None, num_other_drafts

        return draft, num_other_drafts

    def create(
        self,
        resource_id: int,
        base_resource_id: int | None,
        payload: dict[str, Any],
        commit: bool = True,
        **kwargs,
    ) -> tuple[models.DraftResource | None, int]:
        """Create a new draft from an existing resource.

        Args:
            resource_id: The ID of the resource this draft modifies.
            payload: The contents of the draft.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created draft object.

        Raises:
            DraftAlreadyExistsError: If a draft already exists for this resource.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=resource_id, resource_type=self._resource_type, is_deleted=False
        )
        resource = db.session.scalars(stmt).first()

        if resource is None:
            raise EntityDoesNotExistError(self._resource_type, resource_id=resource_id)

        existing_draft, num_other_drafts = self.get(resource_id, log=log)
        if existing_draft:
            raise DraftAlreadyExistsError(self._resource_type, resource_id)

        draft_payload = {
            "resource_data": payload,
            "resource_id": resource_id,
            "resource_snapshot_id": resource.latest_snapshot_id,
            "base_resource_id": base_resource_id,
        }

        draft = models.DraftResource(
            resource_type=self._resource_type,
            target_owner=resource.owner,
            payload=draft_payload,
            creator=current_user,
        )
        db.session.add(draft)

        if commit:
            db.session.commit()
            log.debug("Draft creation successful", resource_id=draft.draft_resource_id)

        return draft, num_other_drafts

    def modify(
        self,
        resource_id: int,
        payload: dict[str, Any],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> tuple[models.DraftResource | None, int]:
        """Modify a draft.

        Args:
            resource_id: The unique id of the resource.
            payload: The contents of the draft.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated draft resource object.

        Raises:
            DraftDoesNotExistError: If the draft is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        draft, num_other_drafts = self.get(
            resource_id, error_if_not_found=error_if_not_found, log=log
        )

        if draft is None:
            return None, num_other_drafts

        current_timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        draft.payload["resource_data"] = payload
        draft.last_modified_on = current_timestamp

        if commit:
            db.session.commit()
            log.debug("Draft modification successful", resource_id=resource_id)

        return draft, num_other_drafts

    def delete(self, resource_id: int, **kwargs) -> dict[str, Any]:
        """Delete a draft.

        Args:
            resource_id: The unique id of the resource.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        draft, _ = cast(
            tuple[models.DraftResource, int],
            self.get(resource_id, error_if_not_found=True, log=log),
        )
        draft_id = draft.draft_resource_id
        db.session.delete(draft)
        db.session.commit()
        log.debug("Draft deleted", resource_id=resource_id)

        return {"status": "Success", "id": [draft_id]}



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Optional, Tuple

import structlog
from structlog.stdlib import BoundLogger

from .conv2dbatchnorm import Conv2DBatchNormalization

LOGGER: BoundLogger = structlog.stdlib.get_logger()


try:
    import tensorflow as tf
    from tensorflow.keras.initializers import Constant, RandomNormal
    from tensorflow.keras.layers import (
        Activation,
        Concatenate,
        Conv2D,
        Dense,
        Dropout,
        Flatten,
        Layer,
        LeakyReLU,
        Reshape,
        Softmax,
    )

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class YOLOV1BoundingBoxCoordinates(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        confidence_indices = [5 * x + 4 for x in range(n_bounding_boxes)]
        self.bounding_box_indices = [
            x for x in range(5 * n_bounding_boxes) if x not in confidence_indices
        ]
        self.logits_reshape = Reshape(
            target_shape=(
                grid_shape[0],
                grid_shape[1],
                n_bounding_boxes,
                4,
            ),
            name=f"{name}_reshape",
        )
        self.concatenate = Concatenate(axis=-1)
        self.xy_activation = Activation(
            "sigmoid",
            name=f"{name}_xy_activation",
        )
        self.wh_activation = Activation(
            "exponential",
            name=f"{name}_wh_activation",
        )

    def call(self, inputs):
        x = tf.gather(inputs, indices=self.bounding_box_indices, axis=-1)
        x = self.logits_reshape(x)
        x_xy = self.xy_activation(x[..., :2])
        x_wh = self.wh_activation(x[..., 2:4])

        return self.concatenate([x_xy, x_wh])

    def get_config(self):
        grid_shape = self.logits_reshape.target_shape[:2]
        n_bounding_boxes = self.logits_reshape.target_shape[2]

        config = super().get_config()
        config.update({"grid_shape": grid_shape, "n_bounding_boxes": n_bounding_boxes})

        return config


class YOLOV1BoundingBoxConfidences(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.confidence_indices = [5 * x + 4 for x in range(n_bounding_boxes)]
        self.logits_reshape = Reshape(
            target_shape=(
                grid_shape[0],
                grid_shape[1],
                n_bounding_boxes,
            ),
            name=f"{name}_reshape",
        )
        self.activation = Activation(
            "sigmoid",
            name=f"{name}_activation",
        )

    def call(self, inputs):
        x = tf.gather(inputs, indices=self.confidence_indices, axis=-1)
        x = self.logits_reshape(x)

        return self.activation(x)

    def get_config(self):
        grid_shape = self.logits_reshape.target_shape[:2]
        n_bounding_boxes = self.logits_reshape.target_shape[2]

        config = super().get_config()
        config.update({"grid_shape": grid_shape, "n_bounding_boxes": n_bounding_boxes})

        return config


class YOLOV1ClassProbabilities(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        n_classes: int,
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.index_begin = 5 * n_bounding_boxes
        self.logits_reshape = Reshape(
            target_shape=(
                grid_shape[0],
                grid_shape[1],
                n_classes,
            ),
            name=f"{name}_reshape",
        )
        self.activation = Activation(
            Softmax(),
            name=f"{name}_activation",
        )

    def call(self, inputs):
        x = inputs[..., self.index_begin :]  # noqa: E203
        x = self.logits_reshape(x)

        return self.activation(x)

    def get_config(self):
        grid_shape = self.logits_reshape.target_shape[:2]
        n_bounding_boxes = self.index_begin // 5
        n_classes = self.logits_reshape.target_shape[2]

        config = super().get_config()
        config.update(
            {
                "grid_shape": grid_shape,
                "n_bounding_boxes": n_bounding_boxes,
                "n_classes": n_classes,
            }
        )

        return config


class YOLOV1Detector(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        n_classes: int,
        init_final_bias_confidence: float = 0.01,
        name: str = "yolo_v1_detector",
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        final_bias_init = tf.cast(
            tf.math.log((1 - init_final_bias_confidence) / init_final_bias_confidence),
            tf.float32,
        )
        self.conv_1 = Conv2DBatchNormalization(
            filters=1024,
            kernel_size=(3, 3),
            strides=(1, 1),
            padding="same",
            name=f"{name}_conv_1",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.conv_2 = Conv2DBatchNormalization(
            filters=1024,
            kernel_size=(3, 3),
            strides=(1, 1),
            padding="same",
            name=f"{name}_conv_2",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.conv_3 = Conv2DBatchNormalization(
            filters=1024,
            kernel_size=(3, 3),
            strides=(1, 1),
            padding="same",
            name=f"{name}_conv_3",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.conv_4 = Conv2DBatchNormalization(
            filters=1024,
            kernel_size=(3, 3),
            strides=(1, 1),
            padding="same",
            name=f"{name}_conv_4",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.flatten = Flatten(name=f"{name}_flatten")
        self.dense_1 = Dense(
            units=4096,
            name=f"{name}_dense_1",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.dense_1_activation = Activation(
            LeakyReLU(alpha=0.1), name=f"{name}_dense_1_activation"
        )
        self.dense_1_dropout = Dropout(0.5, name=f"{name}_dense_1_dropout")
        self.dense_2 = Dense(
            units=grid_shape[0] * grid_shape[1] * (5 * n_bounding_boxes + n_classes),
            name=f"{name}_dense_2",
            kernel_initializer=RandomNormal(stddev=0.01),
            bias_initializer=Constant(final_bias_init),
        )
        self.dense_2_reshape = Reshape(
            target_shape=(
                grid_shape[0],
                grid_shape[1],
                5 * n_bounding_boxes + n_classes,
            ),
            name=f"{name}_dense_2_reshape",
        )
        self.bbox_coordinates = YOLOV1BoundingBoxCoordinates(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            name=f"{name}_bbox_coordinates",
        )
        self.bbox_confidences = YOLOV1BoundingBoxConfidences(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            name=f"{name}_bbox_confidences",
        )
        self.class_probabilities = YOLOV1ClassProbabilities(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            n_classes=n_classes,
            name=f"{name}_class_probabilities",
        )

    def call(self, inputs, training=None):
        x = self.conv_1(inputs, training=training)
        x = self.conv_2(x, training=training)
        x = self.conv_3(x, training=training)
        x = self.conv_4(x, training=training)
        x = self.flatten(x)
        x = self.dense_1(x)
        x = self.dense_1_activation(x)
        x = self.dense_1_dropout(x, training=training)
        x = self.dense_2(x)
        x = self.dense_2_reshape(x)

        bbox_coordinates = self.bbox_coordinates(x)
        bbox_confidences = self.bbox_confidences(x)
        class_probabilities = self.class_probabilities(x)

        return bbox_coordinates, bbox_confidences, class_probabilities

    def get_config(self):
        grid_shape = self.dense_2_reshape.target_shape[:2]
        n_bounding_boxes = self.bbox_coordinates.logits_reshape.target_shape[2]
        n_classes = self.class_probabilities.logits_reshape.target_shape[2]

        config = super().get_config()
        config.update(
            {
                "grid_shape": grid_shape,
                "n_bounding_boxes": n_bounding_boxes,
                "n_classes": n_classes,
            }
        )

        return config


class SimpleYOLOV1Detector(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        n_classes: int,
        init_final_bias_confidence: float = 0.01,
        name: str = "simple_yolo_v1_detector",
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        final_bias_init = tf.cast(
            tf.math.log((1 - init_final_bias_confidence) / init_final_bias_confidence),
            tf.float32,
        )
        self.conv_1 = Conv2D(
            filters=5 * n_bounding_boxes + n_classes,
            kernel_size=(1, 1),
            strides=(1, 1),
            padding="same",
            name=f"{name}_conv_1",
            kernel_initializer=RandomNormal(stddev=0.01),
            bias_initializer=Constant(final_bias_init),
        )
        self.bbox_coordinates = YOLOV1BoundingBoxCoordinates(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            name=f"{name}_bbox_coordinates",
        )
        self.bbox_confidences = YOLOV1BoundingBoxConfidences(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            name=f"{name}_bbox_confidences",
        )
        self.class_probabilities = YOLOV1ClassProbabilities(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            n_classes=n_classes,
            name=f"{name}_class_probabilities",
        )

        self._grid_shape = grid_shape
        self._n_bounding_boxes = n_bounding_boxes
        self._n_classes = n_classes

    def call(self, inputs):
        x = self.conv_1(inputs)

        bbox_coordinates = self.bbox_coordinates(x)
        bbox_confidences = self.bbox_confidences(x)
        class_probabilities = self.class_probabilities(x)

        return bbox_coordinates, bbox_confidences, class_probabilities

    def get_config(self):
        grid_shape = self._grid_shape
        n_classes = self._n_bounding_boxes
        n_bounding_boxes = self._n_classes

        config = super().get_config()
        config.update(
            {
                "grid_shape": grid_shape,
                "n_bounding_boxes": n_bounding_boxes,
                "n_classes": n_classes,
            }
        )

        return config


class BoundingBoxRegressorYOLOV1Head(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        name: str = "bbox_regressor_yolo_v1_head",
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.bbox_conv_1 = Conv2DBatchNormalization(
            filters=1024,
            kernel_size=(1, 1),
            strides=(1, 1),
            padding="valid",
            use_bias=False,
            name=f"{name}_conv_1",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.bbox_conv_1_dropout = Dropout(0.5, name=f"{name}_conv_1_dropout")
        self.bbox_conv_2 = Conv2DBatchNormalization(
            filters=512,
            kernel_size=(1, 1),
            strides=(1, 1),
            padding="valid",
            use_bias=False,
            name=f"{name}_conv_2",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.bbox_conv_2_dropout = Dropout(0.5, name=f"{name}_conv_2_dropout")
        self.bbox_conv_3 = Conv2D(
            filters=5 * n_bounding_boxes,
            kernel_size=(1, 1),
            strides=(1, 1),
            padding="valid",
            use_bias=True,
            name=f"{name}_conv_3",
            kernel_initializer=RandomNormal(stddev=0.01),
        )

        self._grid_shape = grid_shape
        self._n_bounding_boxes = n_bounding_boxes

    def call(self, inputs, training=None):
        x = self.bbox_conv_1(inputs, training=training)
        x = self.bbox_conv_1_dropout(x, training=training)
        x = self.bbox_conv_2(x, training=training)
        x = self.bbox_conv_2_dropout(x, training=training)

        return self.bbox_conv_3(x, training=training)

    def get_config(self):
        grid_shape = self._grid_shape
        n_bounding_boxes = self._n_bounding_boxes

        config = super().get_config()
        config.update({"grid_shape": grid_shape, "n_bounding_boxes": n_bounding_boxes})

        return config


class ObjectClassifierYOLOV1Head(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        n_classes: int,
        init_final_bias_confidence: float = 0.01,
        name: str = "object_classifier_yolo_v1_head",
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        final_bias_init = tf.cast(
            tf.math.log((1 - init_final_bias_confidence) / init_final_bias_confidence),
            tf.float32,
        )
        self.classifier_conv_1 = Conv2DBatchNormalization(
            filters=1024,
            kernel_size=(1, 1),
            strides=(1, 1),
            padding="valid",
            use_bias=False,
            name=f"{name}_conv_1",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.classifier_conv_1_dropout = Dropout(0.5, name=f"{name}_conv_1_dropout")
        self.classifier_conv_2 = Conv2DBatchNormalization(
            filters=512,
            kernel_size=(1, 1),
            strides=(1, 1),
            padding="valid",
            use_bias=False,
            name=f"{name}_conv_2",
            kernel_initializer=RandomNormal(stddev=0.01),
        )
        self.classifier_conv_2_dropout = Dropout(0.5, name=f"{name}_conv_2_dropout")
        self.classifier_conv_3 = Conv2D(
            filters=n_classes,
            kernel_size=(1, 1),
            strides=(1, 1),
            padding="valid",
            use_bias=True,
            name=f"{name}_conv_3",
            kernel_initializer=RandomNormal(stddev=0.01),
            bias_initializer=Constant(final_bias_init),
        )

        self._grid_shape = grid_shape
        self._n_bounding_boxes = n_bounding_boxes
        self._n_classes = n_classes

    def call(self, inputs, training=None):
        x = self.classifier_conv_1(inputs, training=training)
        x = self.classifier_conv_1_dropout(x, training=training)
        x = self.classifier_conv_2(x, training=training)
        x = self.classifier_conv_2_dropout(x, training=training)

        return self.classifier_conv_3(x, training=training)

    def get_config(self):
        grid_shape = self._grid_shape
        n_bounding_boxes = self._n_bounding_boxes
        n_classes = self._n_classes

        config = super().get_config()
        config.update(
            {
                "grid_shape": grid_shape,
                "n_bounding_boxes": n_bounding_boxes,
                "n_classes": n_classes,
            }
        )

        return config


class TwoHeadedYOLOV1Detector(Layer):
    def __init__(
        self,
        grid_shape: Tuple[int, int],
        n_bounding_boxes: int,
        n_classes: int,
        init_final_bias_confidence: float = 0.01,
        name: str = "two_headed_yolo_v1_detector",
        **kwargs,
    ) -> None:
        super().__init__(name=name, **kwargs)
        self.bbox_regressor = BoundingBoxRegressorYOLOV1Head(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            name=f"{name}_bbox_regressor_head",
        )
        self.object_classifier = ObjectClassifierYOLOV1Head(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            n_classes=n_classes,
            init_final_bias_confidence=init_final_bias_confidence,
            name=f"{name}_obj_classifier_head",
        )
        self.concatenate = Concatenate(axis=-1)
        self.bbox_coordinates = YOLOV1BoundingBoxCoordinates(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            name=f"{name}_bbox_coordinates",
        )
        self.bbox_confidences = YOLOV1BoundingBoxConfidences(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            name=f"{name}_bbox_confidences",
        )
        self.class_probabilities = YOLOV1ClassProbabilities(
            grid_shape=grid_shape,
            n_bounding_boxes=n_bounding_boxes,
            n_classes=n_classes,
            name=f"{name}_class_probabilities",
        )

    def call(self, inputs):
        logits_bbox_regressor = self.bbox_regressor(inputs)
        logits_obj_classifier = self.object_classifier(inputs)
        x = self.concatenate([logits_bbox_regressor, logits_obj_classifier])

        bbox_coordinates = self.bbox_coordinates(x)
        bbox_confidences = self.bbox_confidences(x)
        class_probabilities = self.class_probabilities(x)

        return bbox_coordinates, bbox_confidences, class_probabilities

    def get_config(self):
        object_classifier_config = self.object_classifier.get_config()
        grid_shape = object_classifier_config["grid_shape"]
        n_bounding_boxes = object_classifier_config["n_bounding_boxes"]
        n_classes = object_classifier_config["n_classes"]

        config = super().get_config()
        config.update(
            {
                "grid_shape": grid_shape,
                "n_bounding_boxes": n_bounding_boxes,
                "n_classes": n_classes,
            }
        )

        return config



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform user endpoint operations."""
from __future__ import annotations

import datetime
import uuid
from typing import Any, Final, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.db.models.constants import user_lock_types
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    NoCurrentUserError,
    QueryParameterValidationError,
    UserPasswordChangeError,
    UserPasswordError,
)
from dioptra.restapi.v1.groups.service import GroupMemberService, GroupNameService
from dioptra.restapi.v1.plugin_parameter_types.service import (
    BuiltinPluginParameterTypeService,
)
from dioptra.restapi.v1.shared.password_service import PasswordService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

DEFAULT_GROUP_NAME: Final[str] = "public"
DEFAULT_GROUP_PERMISSIONS: Final[dict[str, Any]] = {
    "read": True,
    "write": True,
    "share_read": False,
    "share_write": False,
}
DAYS_TO_EXPIRE_PASSWORD_DEFAULT: Final[int] = 365
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "username": lambda x: models.User.username.like(x, escape="/"),
    "email": lambda x: models.User.email_address.like(x, escape="/"),
}


class UserService(object):
    """The service methods used for creating and managing user accounts."""

    @inject
    def __init__(
        self,
        user_password_service: UserPasswordService,
        user_name_service: UserNameService,
        group_name_service: GroupNameService,
        group_member_service: GroupMemberService,
        builtin_plugin_parameter_type_service: BuiltinPluginParameterTypeService,
    ) -> None:
        """Initialize the user service.

        All arguments are provided via dependency injection.

        Args:
            user_password_service: A UserPasswordService object.
            user_name_service: A UserNameService object.
            group_name_service: A GroupNameService object.
            group_member_service: A GroupMemberService object.
            builtin_plugin_parameter_type_service: A BuiltinPluginParameterTypeService
                object.
        """
        self._user_password_service = user_password_service
        self._user_name_service = user_name_service
        self._group_name_service = group_name_service
        self._group_member_service = group_member_service
        self._builtin_plugin_parameter_type_service = (
            builtin_plugin_parameter_type_service
        )

    def create(
        self,
        username: str,
        email_address: str,
        password: str,
        confirm_password: str,
        commit: bool = True,
        **kwargs,
    ) -> models.User:
        """Create a new user.

        Args:
            username: The username requested by the new user. Must be unique.
            email_address: The email address to associate with the new user. Must be
                unique.
            password: The password for the new user.
            confirm_password: The password confirmation for the new user.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The new user object.

        Raises:
            UserRegistrationError: If the password and confirmation password do not
                match.
            UsernameNotAvailableError: If the username already exists.
            UserEmailNotAvailableError: If the email already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        if password != confirm_password:
            raise QueryParameterValidationError(
                "password", "equivalence", password="***", confirmation="***"
            )

        duplicate = self._user_name_service.get(username, log=log)
        if duplicate is not None:
            raise EntityExistsError("User", duplicate.user_id, username=username)

        duplicate = self._get_user_by_email(email_address, log=log)
        if duplicate is not None:
            raise EntityExistsError(
                "User", duplicate.user_id, email_address=email_address
            )

        hashed_password = self._user_password_service.hash(password, log=log)
        new_user: models.User = models.User(
            username=username, password=hashed_password, email_address=email_address
        )

        default_group = self._create_or_get_default_group(
            user=new_user,
            log=log,
        )
        self._group_member_service.create(
            default_group,
            user=new_user,
            permissions=DEFAULT_GROUP_PERMISSIONS,
            commit=False,
            log=log,
        )

        db.session.add(new_user)
        db.session.add(default_group)

        if commit:
            db.session.commit()
            log.debug("User registration successful", user_id=new_user.user_id)

        return new_user

    def get(
        self,
        search_string: str,
        page_index: int,
        page_length: int,
        **kwargs,
    ) -> tuple[list[models.User], int]:
        """Fetch a list of users, optionally filtering by search string and paging
        parameters.

        Args:
            search_string: A search string used to filter results.
            page_index: The index of the first user to be returned.
            page_length: The maximum number of users to be returned.

        Returns:
            A tuple containing a list of users and the total number of users matching
            the query.

        Raises:
            BackendDatabaseError: If the database query returns a None when counting
                the number of users.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get list of users")

        search_filters = construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)

        stmt = (
            select(func.count(models.User.user_id))
            .filter_by(is_deleted=False)
            .filter(search_filters)
        )
        total_num_users = db.session.scalars(stmt).first()

        if total_num_users is None:
            log.error(
                "The database query returned a None when counting the number of "
                "users when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_users == 0:
            return cast(list[models.User], []), total_num_users

        stmt = (
            select(models.User)  # type: ignore
            .filter_by(is_deleted=False)
            .filter(search_filters)
            .offset(page_index)
            .limit(page_length)
        )
        users = cast(list[models.User], db.session.scalars(stmt).all())

        return users, total_num_users

    def _get_user_by_email(
        self, email_address: str, error_if_not_found: bool = False, **kwargs
    ) -> models.User | None:
        """Lookup a user by email address.

        Args:
            email_address: The email address of the user.
            error_if_not_found: If True, raise an error if the user is not found.
                Defaults to False.

        Returns:
            The user object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the user is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Lookup user account by email", email_address=email_address)

        stmt = select(models.User).filter_by(
            email_address=email_address, is_deleted=False
        )
        user: models.User | None = db.session.scalars(stmt).first()

        if user is None:
            if error_if_not_found:
                raise EntityDoesNotExistError("User", email_address=email_address)

            return None

        return user

    def _create_or_get_default_group(
        self,
        user: models.User,
        **kwargs,
    ) -> models.Group:
        """Returns the default group if it exists, otherwise create and return it.

        Args:
            user: The user to assign as the creator of the group. If group already
                exists, then this does nothing.

        Returns:
            The group object if found, otherwise None.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        if (
            group := self._group_name_service.get(DEFAULT_GROUP_NAME, log=log)
        ) is not None:
            return group

        default_group = models.Group(name=DEFAULT_GROUP_NAME, creator=user)
        # Register the built-in plugin parameter types when creating a new group.
        self._builtin_plugin_parameter_type_service.create_all(
            user=user, group=default_group, commit=False
        )
        return default_group


class UserIdService(object):
    """The service methods used to manage a user by ID."""

    @inject
    def __init__(
        self,
        user_password_service: UserPasswordService,
    ) -> None:
        """Initialize the current user service.

        All arguments are provided via dependency injection.

        Args:
            user_password_service: A UserPasswordService object.
        """
        self._user_password_service = user_password_service

    def get(
        self, user_id: int, error_if_not_found: bool = False, **kwargs
    ) -> models.User | None:
        """Fetch a user by its unique id.

        Args:
            user_id: The unique id of the user.
            error_if_not_found: If True, raise an error if the user is not found.
                Defaults to False.

        Returns:
            The user object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the user is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Lookup user account by unique id", user_id=user_id)

        stmt = select(models.User).filter_by(user_id=user_id, is_deleted=False)
        user = db.session.scalars(stmt).first()

        if user is None:
            if error_if_not_found:
                raise EntityDoesNotExistError("User", user_id=user_id)

            return None

        return user

    def change_password(
        self,
        user_id: int,
        current_password: str,
        new_password: str,
        confirm_new_password: str,
        **kwargs,
    ) -> dict[str, Any]:
        """Change a user's password.

        Args:
            user_id: The user's unique id.
            current_password: The user's current password.
            new_password: The user's new password, to replace the current one after
                authentication.
            confirm_new_password: Confirmation of the new password.

        Returns:
            A dictionary containing the password change success message if the password
            change is successful.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Change a user's password by ID.", user_id=user_id)
        user = cast(
            models.User, self.get(user_id=user_id, error_if_not_found=True, log=log)
        )
        return self._user_password_service.change(
            user=user,
            current_password=current_password,
            new_password=new_password,
            confirm_new_password=confirm_new_password,
            log=log,
        )


class UserCurrentService(object):
    """The service methods used to manage the current user."""

    @inject
    def __init__(
        self,
        user_id_service: UserIdService,
        user_password_service: UserPasswordService,
    ) -> None:
        """Initialize the current user service.

        All arguments are provided via dependency injection.

        Args:
            user_id_service: A UserIdService object.
            user_password_service: A UserPasswordService object.
        """
        self._user_id_service = user_id_service
        self._user_password_service = user_password_service

    def get(self, **kwargs) -> models.User:
        """Fetch information about the current user.

        Returns:
            The current user object.

        Raises:
            NoCurrentUserError: If there is no current user.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())  # noqa: F841

        if not current_user.is_authenticated:
            raise NoCurrentUserError

        return cast(models.User, current_user)

    def modify(
        self, username: str, email_address: str, commit: bool = True, **kwargs
    ) -> models.User:
        """Modifies the current user

        Args:
            username: The user's current username.
            email_address: The user's current email_address.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The current user object.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Modify user account", user_id=current_user.user_id)

        current_timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        current_user.username = username
        current_user.email_address = email_address
        current_user.last_modified_on = current_timestamp

        if commit:
            db.session.commit()

        return cast(models.User, current_user)

    def delete(self, password: str, **kwargs) -> dict[str, Any]:
        """Permanently deletes the current user.

        Args:
            password: The current user's password.

        Returns:
            A dictionary containing the delete user success message if the user is
            deleted successfully.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Delete user account", user_id=current_user.user_id)
        self._user_password_service.authenticate(
            password=password,
            user_password=current_user.password,
            expiration_date=current_user.password_expire_on,
            error_if_failed=True,
            log=log,
        )

        user_id = current_user.user_id
        username = current_user.username

        deleted_user_lock = models.UserLock(
            user_lock_type=user_lock_types.DELETE,
            user=current_user,
        )
        db.session.add(deleted_user_lock)
        db.session.commit()
        log.debug("User account deleted", user_id=user_id, username=username)

        return {"status": "Success", "id": [user_id]}

    def change_password(
        self,
        current_password: str,
        new_password: str,
        confirm_new_password: str,
        **kwargs,
    ) -> dict[str, Any]:
        """Change the current user's password.

        Args:
            current_password: The user's current password.
            new_password: The user's new password, to replace the current one after
                authentication.
            confirm_new_password: Confirmation of the new password.

        Returns:
            A dictionary containing the password change success message if the password
            change is successful.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Change the current user's password", user_id=current_user.user_id)
        return self._user_password_service.change(
            user=current_user,
            current_password=current_password,
            new_password=new_password,
            confirm_new_password=confirm_new_password,
            log=log,
        )


class UserNameService(object):
    """The service methods used to register and manage user accounts by username."""

    @inject
    def __init__(
        self,
        user_password_service: UserPasswordService,
    ) -> None:
        """Initialize the user name service.

        All arguments are provided via dependency injection.

        Args:
            user_password_service: A UserPasswordService object.
        """
        self._user_password_service = user_password_service

    def get(
        self, username: str, error_if_not_found: bool = False, **kwargs
    ) -> models.User | None:
        """Fetch a user by its username.

        Args:
            username: The username of the user.
            error_if_not_found: If True, raise an error if the user is not found.
                Defaults to False.

        Returns:
            The user object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the user is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Lookup user account by unique username", username=username)

        stmt = select(models.User).filter_by(username=username, is_deleted=False)
        user = db.session.scalars(stmt).first()

        if user is None:
            if error_if_not_found:
                raise EntityDoesNotExistError("User", username=username)

            return None

        return user


class UserPasswordService(object):
    """The service methods used to manage user passwords."""

    @inject
    def __init__(
        self,
        password_service: PasswordService,
    ) -> None:
        """Initialize the user password service.

        All arguments are provided via dependency injection.

        Args:
            password_service: A PasswordService object.
        """
        self._password_service = password_service

    def authenticate(
        self,
        password: str,
        user_password: str,
        expiration_date: datetime.datetime,
        error_if_failed: bool = False,
        **kwargs,
    ) -> bool:
        """Authenticate a user's password.

        Args:
            password: The password to verify.
            user_password: The user's hashed password.
            expiration_date: The date and time the user's password is set to expire.
            error_if_failed: If True, raise an error if the password verification fails.
                Defaults to False.

        Returns:
            True if the password is verified, otherwise False.

        Raises:
            UserPasswordVerificationError: If the password verification fails and
                `error_if_failed` is True.
            UserPasswordExpiredError: If the user's password has expired.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        current_timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        authenticated = self._password_service.verify(
            password=password, hashed_password=user_password, log=log
        )

        if not authenticated and error_if_failed:
            raise UserPasswordError("Password authentication failed.")

        if expiration_date < current_timestamp:
            raise UserPasswordError("Password expired.")

        return authenticated

    def change(
        self,
        user: models.User,
        current_password: str,
        new_password: str,
        confirm_new_password: str,
        commit: bool = True,
        **kwargs,
    ) -> dict[str, Any]:
        """Change a user's password.

        Args:
            user: The user object.
            current_password: The user's current password.
            new_password: The user's new password, to replace the current one after
                authentication.
            confirm_new_password: Confirmation of the new password.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            A dictionary containing the password change success message if the password
            change is successful.

        Raises:
            UserPasswordChangeError: If the password change fails.
            UserPasswordChangeSamePasswordError: If the new password is the same as the
                current password.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        if not self._password_service.verify(
            password=current_password, hashed_password=str(user.password), log=log
        ):
            raise UserPasswordChangeError("Invalid Current Password.")

        if new_password != confirm_new_password:
            raise UserPasswordChangeError(
                "Confirmation password does not match new password."
            )

        if self._password_service.verify(
            password=new_password, hashed_password=str(user.password), log=log
        ):
            raise UserPasswordChangeError("New password matches old password.")

        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        user.password = self._password_service.hash(password=new_password, log=log)
        user.alternative_id = uuid.uuid4()
        user.last_modified_on = timestamp
        user.password_expire_on = timestamp + datetime.timedelta(
            days=DAYS_TO_EXPIRE_PASSWORD_DEFAULT
        )

        if commit:
            db.session.commit()

        return {"status": "Password Change Success", "username": user.username}

    def hash(self, password: str, **kwargs) -> str:
        """Hash a password.

        Args:
            password: The password to hash.

        Returns:
            The hashed password.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        return self._password_service.hash(password=password, log=log)


def load_user(user_id: str) -> models.User | None:
    """Load the user associated with a provided id.

    This function is intended for use with Flask-Login. The use of the alternative ID is
    needed to support the "logout everywhere" functionality and provides a mechanism for
    expiring sessions.


    Args:
        user_id: A string containing a UUID that matches the user's alternative ID.

    Returns:
        A user object if the user is found, otherwise None.
    """
    stmt = select(models.User).filter_by(
        alternative_id=uuid.UUID(user_id), is_deleted=False
    )
    return db.session.scalars(stmt).first()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the auth endpoints."""
from __future__ import annotations

import uuid
from typing import Any

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from .schema import LoginSchema, LogoutQueryParametersSchema, LogoutSchema
from .service import AuthService

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Authentication", description="Authentication endpoint")


@api.route("/login")
class LoginResource(Resource):
    """Methods for the /auth/login endpoint."""

    @inject
    def __init__(self, auth_service: AuthService, *args, **kwargs) -> None:
        """Initialize the login resource.

        All arguments are provided via dependency injection.

        Args:
            auth_service: An AuthService object.
        """
        self._auth_service = auth_service
        super().__init__(*args, **kwargs)

    @accepts(schema=LoginSchema, api=api)
    @responds(schema=LoginSchema, api=api)
    def post(self) -> dict[str, Any]:
        """Login to a registered user account."""
        log: BoundLogger = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="auth/login", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore
        return self._auth_service.login(
            username=parsed_obj["username"], password=parsed_obj["password"], log=log
        )


@api.route("/logout")
class LogoutResource(Resource):
    """Methods for the /auth/logout endpoint."""

    @inject
    def __init__(self, auth_service: AuthService, *args, **kwargs) -> None:
        """Initialize the login resource.

        All arguments are provided via dependency injection.

        Args:
            auth_service: An AuthService object.
        """
        self._auth_service = auth_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=LogoutQueryParametersSchema, api=api)
    @responds(schema=LogoutSchema, api=api)
    def post(self) -> dict[str, Any]:
        """Logout the current user.

        Must be logged in.
        """
        log: BoundLogger = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="auth/logout", request_type="POST"
        )
        parsed_query_params = request.parsed_query_params  # type: ignore
        return self._auth_service.logout(
            everywhere=parsed_query_params["everywhere"], log=log
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
A CLI tool used to start up a worker process for running jobs.
"""

import logging
import os
import sys

import rq.cli

from dioptra.sdk.utilities.logging import (
    attach_stdout_stream_handler,
    set_logging_level,
)

_REQUIRED_ENV = {
    "MLFLOW_S3_ENDPOINT_URL",
    "DIOPTRA_API",
    "DIOPTRA_WORKER_USERNAME",
    "DIOPTRA_WORKER_PASSWORD",
}


def _setup_logging() -> None:
    attach_stdout_stream_handler(
        True if os.getenv("DIOPTRA_RQ_WORKER_LOG_AS_JSON") else False,
    )
    set_logging_level(os.getenv("DIOPTRA_RQ_WORKER_LOG_LEVEL", default="INFO"))


def main() -> int:
    _setup_logging()
    log = logging.getLogger("dioptra-worker")
    exit_status = 0

    # We know what functions will be executed through rq and what they
    # require, so we may as well check that before starting up the worker.
    # Better to error out as early as possible.
    unset_vars = _REQUIRED_ENV - os.environ.keys()
    if unset_vars:
        exit_status = 1
        log.fatal("Environment variables must be set: %s", ", ".join(unset_vars))
    else:
        # Disabling standalone mode means the worker function can return a
        # value to this script, exceptions will propagate to us, etc. (as
        # opposed to being intercepted and handled by click).  As a wrapper
        # that seems appropriate, although I don't think the worker function
        # is written to return anything, and we presently don't need to
        # specially handle any of the exceptions.
        rq.cli.worker(standalone_mode=False)

    return exit_status


if __name__ == "__main__":
    sys.exit(main())



"""Add index over the draft_resources.payload["resource_id"] JSON field.

Revision ID: b5cc97666e3d
Revises: 148c032ffbd8
Create Date: 2024-06-20 17:10:13.659783

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b5cc97666e3d"
down_revision = "148c032ffbd8"
branch_labels = None
depends_on = None

POSTGRES_DIALECT = "postgresql"
SQLITE_DIALECT = "sqlite"


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    dialect_name = op.get_context().dialect.name

    with op.batch_alter_table("draft_resources", schema=None) as batch_op:
        if dialect_name == POSTGRES_DIALECT:
            batch_op.create_index(
                batch_op.f("ix_draft_resources_payload_resource_id"),
                [
                    sa.text(
                        "CAST(CAST(payload ->> 'resource_id' AS VARCHAR) AS INTEGER)"
                    )
                ],
                unique=False,
            )

        elif dialect_name == SQLITE_DIALECT:
            batch_op.create_index(
                batch_op.f("ix_draft_resources_payload_resource_id"),
                [
                    sa.text(
                        "CAST(JSON_EXTRACT(payload, '$.\"resource_id\"') AS INTEGER)"
                    )
                ],
                unique=False,
            )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("draft_resources", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_draft_resources_payload_resource_id"))

    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
Test building types from type definitions.
"""
from collections.abc import Mapping
from typing import Any

import pytest

from dioptra.sdk.exceptions.task_engine import (
    AnonymousSimpleTypeError,
    BuiltinTypeRedefinitionError,
    DioptraTypeError,
    InvalidKeyTypeError,
    NonSimpleSuperTypeError,
    TooManyTypeStructuresError,
    TypeNotFoundError,
    TypeReferenceCycleError,
)
from dioptra.task_engine.type_registry import (
    TYPE_ANY,
    TYPE_BOOLEAN,
    TYPE_INTEGER,
    TYPE_NULL,
    TYPE_NUMBER,
    TYPE_STRING,
    build_type_registry,
)
from dioptra.task_engine.types import (
    SimpleType,
    StructuredType,
    StructureType,
    TypeStructure,
    UnionType,
)


def test_builtin_types() -> None:
    types: dict = {}

    type_reg = build_type_registry(types)

    string_type = type_reg["string"]
    int_type = type_reg["integer"]
    number_type = type_reg["number"]
    bool_type = type_reg["boolean"]
    null_type = type_reg["null"]
    any_type = type_reg["any"]

    assert string_type == TYPE_STRING
    assert int_type == TYPE_INTEGER
    assert number_type == TYPE_NUMBER
    assert bool_type == TYPE_BOOLEAN
    assert null_type == TYPE_NULL
    assert any_type == TYPE_ANY

    assert int_type.is_subtype_of(number_type)


def test_build_simple_type() -> None:
    types = {"A": None, "B": {"is_a": "A"}}

    type_reg = build_type_registry(types)

    a_type = type_reg["A"]
    b_type = type_reg["B"]

    assert a_type == SimpleType("A")
    assert b_type == SimpleType("B", super_type=a_type)

    assert b_type.is_subtype_of(a_type)


def test_build_list_type() -> None:
    types = {"A": None, "list_of_a": {"list": "A"}}

    type_reg = build_type_registry(types)

    a_type = type_reg["A"]
    list_of_a_type = type_reg["list_of_a"]

    assert a_type == SimpleType("A")
    assert list_of_a_type == StructuredType(
        TypeStructure(StructureType.LIST, a_type), name="list_of_a"
    )


def test_build_tuple_type() -> None:
    types = {"A": None, "B": None, "tuple_a_b": {"tuple": ["A", "B"]}}

    type_reg = build_type_registry(types)

    a_type = type_reg["A"]
    b_type = type_reg["B"]
    tuple_a_b_type = type_reg["tuple_a_b"]

    assert tuple_a_b_type == StructuredType(
        TypeStructure(StructureType.TUPLE, [a_type, b_type]), name="tuple_a_b"
    )


def test_build_mapping_type_enum() -> None:
    types = {"A": None, "B": None, "mapping_a_b": {"mapping": {"foo": "A", "bar": "B"}}}

    type_reg = build_type_registry(types)

    a_type = type_reg["A"]
    b_type = type_reg["B"]
    mapping_a_b_type = type_reg["mapping_a_b"]

    assert mapping_a_b_type == StructuredType(
        TypeStructure(StructureType.MAPPING, {"foo": a_type, "bar": b_type}),
        name="mapping_a_b",
    )


def test_build_mapping_type_key_value() -> None:
    types = {
        "A": None,
        "mapping_string_a": {"mapping": ["string", "A"]},
        "mapping_integer_a": {"mapping": ["integer", "A"]},
    }

    type_reg = build_type_registry(types)

    a_type = type_reg["A"]
    mapping_string_a_type = type_reg["mapping_string_a"]
    mapping_integer_a_type = type_reg["mapping_integer_a"]

    expected_mapping_string_a_type = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type]),
        name="mapping_string_a",
    )

    expected_mapping_integer_a_type = StructuredType(
        TypeStructure(StructureType.MAPPING, [TYPE_INTEGER, a_type]),
        name="mapping_integer_a",
    )

    assert mapping_string_a_type == expected_mapping_string_a_type
    assert mapping_integer_a_type == expected_mapping_integer_a_type


def test_build_union_type() -> None:
    types = {"A": None, "B": None, "union_a_b": {"union": ["A", "B"]}}

    type_reg = build_type_registry(types)

    a_type = type_reg["A"]
    b_type = type_reg["B"]
    union_a_b = type_reg["union_a_b"]

    assert union_a_b == UnionType([a_type, b_type], name="union_a_b")


def test_build_nested_type() -> None:
    types = {
        "A": None,
        "nested_type": {"mapping": ["string", {"list": {"tuple": ["A", "number"]}}]},
    }

    type_reg = build_type_registry(types)
    a_type = type_reg["A"]
    nested_type = type_reg["nested_type"]

    expected_type = StructuredType(
        TypeStructure(
            StructureType.MAPPING,
            [
                TYPE_STRING,
                StructuredType(
                    TypeStructure(
                        StructureType.LIST,
                        StructuredType(
                            TypeStructure(StructureType.TUPLE, [a_type, TYPE_NUMBER])
                        ),
                    )
                ),
            ],
        ),
        name="nested_type",
    )

    # This will not actually check structure, since both types are named.
    assert nested_type == expected_type

    # This will check structure, and all sub-structure in this case uses
    # anonymous types, so it's a better test of processing complex nested
    # structures/types.
    assert nested_type.structure == expected_type.structure


def test_build_mapping_type_invalid_key_type() -> None:
    types = {"A": None, "bad_mapping_type": {"mapping": ["A", "integer"]}}

    # The key type in a key/value type style mapping type must be either
    # string or integer.
    with pytest.raises(InvalidKeyTypeError) as e:
        build_type_registry(types)

    assert e.value.invalid_key_type == SimpleType("A")


def test_too_many_structure_types() -> None:
    types = {"A": None, "bad": {"list": "A", "tuple": ["A", "number"]}}

    with pytest.raises(TooManyTypeStructuresError) as e:
        build_type_registry(types)

    assert e.value.structure_types == [StructureType.LIST, StructureType.TUPLE]


def test_non_simple_has_supertype() -> None:
    types = {
        "A": None,
        "B": None,
        "bad_list": {
            "list": "A",
            "is_a": "B",  # invalid: structured types can't have a supertype
        },
    }

    with pytest.raises(DioptraTypeError):
        build_type_registry(types)


def test_union_and_structured() -> None:
    types = {
        "A": None,
        "B": None,
        "bad_type": {
            "union": ["A", "B"],
            "tuple": ["A", "B"],  # invalid: a type can't be both union and
        },  # structured
    }

    with pytest.raises(DioptraTypeError):
        build_type_registry(types)


def test_non_simple_supertype() -> None:
    types = {
        "A": None,
        "list_of_a": {"list": "A"},
        "bad_simple": {
            # invalid: supertypes must be simple
            "is_a": "list_of_a"
        },
    }

    with pytest.raises(NonSimpleSuperTypeError) as e:
        build_type_registry(types)

    assert e.value.super_type_name == "list_of_a"


def test_type_not_found() -> None:
    types = {"list_of_a": {"list": "A"}}

    with pytest.raises(TypeNotFoundError) as e:
        build_type_registry(types)

    assert e.value.type_name == "A"


def test_type_reference_cycle() -> None:
    types: Mapping[str, Any] = {"A": {"is_a": "A"}}

    with pytest.raises(TypeReferenceCycleError) as e:
        build_type_registry(types)

    assert e.value.cycle == ["A", "A"]

    types = {
        "A": {"list": "B"},
        "B": {"tuple": ["string", "C"]},
        "C": {"mapping": {"foo": "A"}},
    }

    with pytest.raises(TypeReferenceCycleError) as e:
        build_type_registry(types)

    assert e.value.cycle == ["A", "C", "B", "A"]


def test_redefine_builtin_type() -> None:
    types = {"A": None, "integer": {"list": "A"}}

    with pytest.raises(BuiltinTypeRedefinitionError) as e:
        build_type_registry(types)

    assert e.value.builtin_type_name == "integer"


def test_anonymous_simple_type() -> None:
    types = {
        # In normal use, this would actually produce a semantic validation
        # error due to the non-string key, before we ever tried to build a type
        # out of it.
        None: {}
    }

    with pytest.raises(AnonymousSimpleTypeError):
        build_type_registry(types)

    types = {
        None: {
            # Since there is no union or structure detected in this definition,
            # it is considered a simple type by build_type_registry().  In
            # reality, this would not pass schema validation.
            "foo": 1
        }
    }

    with pytest.raises(AnonymousSimpleTypeError):
        build_type_registry(types)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform auth endpoint operations."""
from __future__ import annotations

import datetime
import uuid
from typing import Any, cast

import structlog
from flask_login import current_user, login_user, logout_user
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.v1.users.service import UserNameService, UserPasswordService

LOGGER: BoundLogger = structlog.stdlib.get_logger()


class AuthService(object):
    """The service methods for user logins and logouts."""

    @inject
    def __init__(
        self,
        user_name_service: UserNameService,
        user_password_service: UserPasswordService,
    ) -> None:
        """Initialize the authentication service.

        All arguments are provided via dependency injection.

        Args:
            user_name_service: A UserNameService object.
            user_password_service: A UserPasswordService object.
        """
        self._user_name_service = user_name_service
        self._user_password_service = user_password_service

    def login(
        self,
        username: str,
        password: str,
        **kwargs,
    ) -> dict[str, Any]:
        """Login the user with the given username and password.

        Args:
            username: The username for logging into the user account.
            password: The password for authenticating the user account.

        Returns:
            A dictionary containing the login success message.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        user = cast(
            models.User,
            self._user_name_service.get(
                username=username, error_if_not_found=True, log=log
            ),
        )
        self._user_password_service.authenticate(
            password=password,
            user_password=str(user.password),
            expiration_date=user.password_expire_on,
            error_if_failed=True,
            log=log,
        )
        login_user(user, remember=True)
        user.last_login_on = datetime.datetime.now(tz=datetime.timezone.utc)
        db.session.commit()
        log.debug("Login successful", user_id=user.user_id)
        return {"status": "Login successful", "username": username}

    def logout(self, everywhere: bool, **kwargs) -> dict[str, Any]:
        """Log the current user out.

        Args:
            everywhere: If True, log out from all devices by regenerating the current
                user's alternative id.

        Returns:
            A dictionary containing the logout success message.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        user_id = current_user.user_id
        username = current_user.username

        if everywhere:
            current_user.alternative_id = uuid.uuid4()
            db.session.commit()

        logout_user()
        log.debug("Logout successful", user_id=user_id, everywhere=everywhere)
        return {
            "status": "Logout successful",
            "username": username,
            "everywhere": everywhere,
        }



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing User resources."""
from marshmallow import Schema, fields

from dioptra.restapi.v1.groups.schema import GroupRefSchema
from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
)


class UserRefSchema(Schema):
    """The reference schema for the data stored in a User resource."""

    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the User resource."),
        dump_only=True,
    )
    username = fields.String(
        attribute="username",
        metadata=dict(description="Username of the User resource."),
    )
    url = fields.Url(
        attribute="url",
        metadata=dict(description="URL for accessing the full User resource."),
        relative=True,
    )


class UserMutableFieldsSchema(Schema):
    """The schema for the mutable data fields in a User resource."""

    username = fields.String(
        attribute="username", metadata=dict(description="Username of the User.")
    )
    email = fields.String(
        attribute="email", metadata=dict(description="Email of the User.")
    )


class UserSchema(UserMutableFieldsSchema):
    """The schema for a User that is not the logged-in User."""

    password = fields.String(
        attribute="password",
        metadata=dict(description="Password for the User resource."),
        load_only=True,
    )
    confirmPassword = fields.String(
        attribute="confirm_password",
        metadata=dict(
            description="The password confirmation when creating a new user account."
        ),
        load_only=True,
    )
    id = fields.Integer(
        attribute="id",
        metadata=dict(description="ID for the User resource."),
        dump_only=True,
    )


class UserCurrentSchema(UserSchema):
    """The schema for the currently logged-in User."""

    groups = fields.Nested(
        GroupRefSchema,
        attribute="groups",
        metadata=dict(description="A list of Groups the User is a part of."),
        many=True,
        dump_only=True,
    )
    createdOn = fields.DateTime(
        attribute="created_on",
        metadata=dict(
            description="Timestamp when the User resource was created.",
            dump_only=True,
        ),
    )
    lastModifiedOn = fields.DateTime(
        attribute="last_modified_on",
        metadata=dict(
            description="Timestamp when the User resource was last modified."
        ),
        dump_only=True,
    )
    lastLoginOn = fields.DateTime(
        attribute="last_login_on",
        metadata=dict(
            description="Timestamp when the User resource was last logged in."
        ),
        dump_only=True,
    )
    passwordExpiresOn = fields.DateTime(
        attribute="password_expires_on",
        metadata=dict(
            description="Timestamp when the User resource password expires on."
        ),
        dump_only=True,
    )


class UserPasswordSchema(Schema):
    """The schema for changing a User's password."""

    oldPassword = fields.String(
        attribute="old_password",
        metadata=dict(description="Old password for the User resource."),
    )
    newPassword = fields.String(
        attribute="new_password",
        metadata=dict(description="New password for the User resource."),
    )
    confirmNewPassword = fields.String(
        attribute="confirm_new_password",
        metadata=dict(
            description="The new password confirmation when changing a password."
        ),
        load_only=True,
    )


class UserDeleteSchema(Schema):
    """The schema for deleting a User."""

    password = fields.String(
        attribute="password",
        metadata=dict(description="The users current password."),
    )


class UserPageSchema(BasePageSchema):
    """The paged schema for the data stored in a User resource."""

    data = fields.Nested(
        UserSchema,
        many=True,
        metadata=dict(description="List of User resources in the current page."),
    )


class UserGetQueryParameters(
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
):
    """The query parameters for the GET method of the /users endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Tags."""
from __future__ import annotations

import uuid

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import ClassAssistedBuilder, inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdListSchema, IdStatusResponseSchema
from dioptra.restapi.v1.tags.schema import TagRefSchema

from .service import ResourceTagsIdService, ResourceTagsService

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Tags", description="Tags sub-endpoint")


def generate_resource_tags_endpoint(
    api: Namespace,
    resource_name: str,
) -> Resource:
    """
    Generates an Endpoint for managing Tags on Resources

    Args:
        api: The API namespace
        resource_name: The name of the resource

    Returns:
        The generated Endpoint class
    """

    @api.route("/<int:id>/tags")
    @api.param("id", f"ID for the {resource_name}.")
    class ResourcesTagsEndpoint(Resource):
        @inject
        def __init__(
            self,
            tag_service: ClassAssistedBuilder[ResourceTagsService],
            *args,
            **kwargs,
        ) -> None:
            self._tag_service = tag_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=TagRefSchema(many=True), api=api)
        def get(self, id: int):
            """Gets the list of Tags for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="GET"
            )
            tags = self._tag_service.get(id, log=log)
            return utils.build_tag_ref_list(tags)

        @login_required
        @accepts(schema=IdListSchema, api=api)
        @responds(schema=TagRefSchema(many=True), api=api)
        def post(self, id: int):
            """Appends one or more Tags to the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # type: ignore
            tags = self._tag_service.append(
                id, tag_ids=parsed_obj["ids"], error_if_not_found=True, log=log
            )
            return utils.build_tag_ref_list(tags)

        @login_required
        @accepts(schema=IdListSchema, api=api)
        @responds(schema=TagRefSchema(many=True), api=api)
        def put(self, id: int):
            """Replaces one or more Tags to the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="POST"
            )
            parsed_obj = request.parsed_obj  # type: ignore
            tags = self._tag_service.modify(
                id, tag_ids=parsed_obj["ids"], error_if_not_found=True, log=log
            )
            return utils.build_tag_ref_list(tags)

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, id: int):
            """Removes all Tags from the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="DELETE"
            )
            return self._tag_service.delete(id, error_if_not_found=True, log=log)

    return ResourcesTagsEndpoint


def generate_resource_tags_id_endpoint(
    api: Namespace,
    resource_name: str,
) -> Resource:
    """
    Generates an Endpoint for removing a Tag from a Resource.

    Args:
        api: The API namespace
        resource_name: The name of the resource

    Returns:
        The generated Endpoint class
    """

    @api.route("/<int:id>/tags/<int:tagId>")
    @api.param("id", f"ID for the {resource_name}.")
    @api.param("tagId", "ID for the Tag.")
    class ResourcesTagsIdEndpoint(Resource):
        @inject
        def __init__(
            self,
            tag_id_service: ClassAssistedBuilder[ResourceTagsIdService],
            *args,
            **kwargs,
        ) -> None:
            self._tag_id_service = tag_id_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, id: int, tagId: int):
            """Removes a Tag from the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="GET"
            )
            return self._tag_id_service.delete(id, tagId, log=log)

    return ResourcesTagsIdEndpoint


def generate_nested_resource_tags_endpoint(
    api: Namespace,
    resource_name: str,
    resource_route: str,
    base_resource_name: str,
) -> Resource:
    """
    Generates an Endpoint for managing Tags on Resources

    Args:
        api: The API namespace
        resource_name: The name of the nested resource
        resource_route: The route to the nested resource
        base_resource_name: The name of the resource

    Returns:
        The generated Endpoint class
    """

    route_singular = resource_route[:-1]
    resource_id = f"{route_singular}Id"

    @api.route(f"/<int:id>/{resource_route}/<int:{resource_id}>/tags/")
    @api.param("id", f"ID for the {base_resource_name}.")
    @api.param(f"{resource_id}", f"ID for the {resource_name}.")
    class ResourcesTagsEndpoint(Resource):
        @inject
        def __init__(
            self,
            tag_service: ClassAssistedBuilder[ResourceTagsService],
            *args,
            **kwargs,
        ) -> None:
            self._tag_service = tag_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=TagRefSchema(many=True), api=api)
        def get(self, id: int, **kwargs):
            """Gets the list of Tags for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="GET"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            tags = self._tag_service.get(kwargs[resource_id], log=log)
            return utils.build_tag_ref_list(tags)

        @login_required
        @accepts(schema=IdListSchema, api=api)
        @responds(schema=TagRefSchema(many=True), api=api)
        def post(self, id: int, **kwargs):
            """Appends one or more Tags to the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="POST"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            parsed_obj = request.parsed_obj  # type: ignore
            tags = self._tag_service.append(
                kwargs[resource_id],
                tag_ids=parsed_obj["ids"],
                error_if_not_found=True,
                log=log,
            )
            return utils.build_tag_ref_list(tags)

        @login_required
        @accepts(schema=IdListSchema, api=api)
        @responds(schema=TagRefSchema(many=True), api=api)
        def put(self, id: int, **kwargs):
            """Replaces one or more Tags to the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="POST"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            parsed_obj = request.parsed_obj  # type: ignore
            tags = self._tag_service.modify(
                kwargs[resource_id],
                tag_ids=parsed_obj["ids"],
                error_if_not_found=True,
                log=log,
            )
            return utils.build_tag_ref_list(tags)

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, id: int, **kwargs):
            """Removes all Tags from the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="DELETE"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            return self._tag_service.delete(
                kwargs[resource_id], error_if_not_found=True, log=log
            )

    return ResourcesTagsEndpoint


def generate_nested_resource_tags_id_endpoint(
    api: Namespace,
    resource_name: str,
    resource_route: str,
    base_resource_name: str,
) -> Resource:
    """
    Generates an Endpoint for removing a Tag from a Resource.

    Args:
        api: The API namespace
        base_resource_name: The name of the resource
        resource_name: The name of the nested resource
        resource_route: The route to the nested resource

    Returns:
        The generated Endpoint class
    """

    route_singular = resource_route[:-1]
    resource_id = f"{route_singular}Id"

    @api.route(f"/<int:id>/{resource_route}/<int:{resource_id}>/tags/<int:tagId>")
    @api.param("id", f"ID for the {base_resource_name}.")
    @api.param(f"{resource_id}", f"ID for the {resource_name}.")
    @api.param("tagId", "ID for the Tag.")
    class ResourcesTagsIdEndpoint(Resource):
        @inject
        def __init__(
            self,
            tag_id_service: ClassAssistedBuilder[ResourceTagsIdService],
            *args,
            **kwargs,
        ) -> None:
            self._tag_id_service = tag_id_service.build(resource_type=resource_name)
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=IdStatusResponseSchema, api=api)
        def delete(self, id: int, tagId: int, **kwargs):
            """Removes a Tag from the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Tag", request_type="DELETE"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            return self._tag_id_service.delete(kwargs[resource_id], tagId, log=log)

    return ResourcesTagsIdEndpoint



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform entrypoint endpoint operations."""
from __future__ import annotations

from typing import Any, Final, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.db.models.constants import resource_lock_types
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    QueryParameterNotUniqueError,
    SortParameterValidationError,
)
from dioptra.restapi.utils import find_non_unique
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.plugins.service import PluginIdsService
from dioptra.restapi.v1.queues.service import RESOURCE_TYPE as QUEUE_RESOURCE_TYPE
from dioptra.restapi.v1.queues.service import QueueIdsService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()
PLUGIN_RESOURCE_TYPE: Final[str] = "entry_point_plugin"

RESOURCE_TYPE: Final[str] = "entry_point"
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.EntryPoint.name.like(x, escape="/"),
    "description": lambda x: models.EntryPoint.description.like(x, escape="/"),
    "task_graph": lambda x: models.EntryPoint.task_graph.like(x, escape="/"),
}
SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "name": models.EntryPoint.name,
    "createdOn": models.EntryPoint.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.EntryPoint.description,
}


class EntrypointService(object):
    """The service methods for creating and managing entrypoints."""

    @inject
    def __init__(
        self,
        entrypoint_name_service: EntrypointNameService,
        plugin_ids_service: PluginIdsService,
        queue_ids_service: QueueIdsService,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the entrypoint service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_name_service: A EntrypointNameService object.
            plugin_ids_service: A PluginIdsService object.
            queue_ids_service: A QueueIdsService object.
            group_id_service: A GroupIdService object.
        """
        self._entrypoint_name_service = entrypoint_name_service
        self._plugin_ids_service = plugin_ids_service
        self._queue_ids_service = queue_ids_service
        self._group_id_service = group_id_service

    def create(
        self,
        name: str,
        description: str,
        task_graph: str,
        parameters: list[dict[str, Any]],
        plugin_ids: list[int],
        queue_ids: list[int],
        group_id: int,
        commit: bool = True,
        **kwargs,
    ) -> utils.EntrypointDict:
        """Create a new entrypoint.

        Args:
            name: The name of the entrypoint. The combination of name and group_id must
                be unique.
            description: The description of the entrypoint.
            group_id: The group that will own the entrypoint.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created entrypoint object.

        Raises:
            EntityExistsError: If a entrypoint with the given name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._entrypoint_name_service.get(name, group_id=group_id, log=log)
        if duplicate is not None:
            raise EntityExistsError(
                RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
            )

        group = self._group_id_service.get(group_id, error_if_not_found=True)
        queues = self._queue_ids_service.get(queue_ids, error_if_not_found=True)
        plugins = self._plugin_ids_service.get(plugin_ids, error_if_not_found=True)

        resource = models.Resource(resource_type=RESOURCE_TYPE, owner=group)

        entrypoint_parameters = [
            models.EntryPointParameter(
                name=param["name"],
                default_value=param["default_value"],
                parameter_type=param["parameter_type"],
                parameter_number=i,
            )
            for i, param in enumerate(parameters)
        ]

        new_entrypoint = models.EntryPoint(
            name=name,
            description=description,
            task_graph=task_graph,
            parameters=entrypoint_parameters,
            resource=resource,
            creator=current_user,
        )

        entry_point_plugin_files = [
            models.EntryPointPluginFile(
                entry_point=new_entrypoint,
                plugin=plugin["plugin"],
                plugin_file=plugin_file,
            )
            for plugin in plugins
            for plugin_file in plugin["plugin_files"]
        ]
        new_entrypoint.entry_point_plugin_files = entry_point_plugin_files

        plugin_resources = [plugin["plugin"].resource for plugin in plugins]
        queue_resources = [queue.resource for queue in queues]
        new_entrypoint.children.extend(plugin_resources + queue_resources)

        db.session.add(new_entrypoint)

        if commit:
            db.session.commit()
            log.debug(
                "Entrypoint registration successful",
                entrypoint_id=new_entrypoint.resource_id,
                name=new_entrypoint.name,
            )

        return utils.EntrypointDict(
            entry_point=new_entrypoint, queues=queues, has_draft=False
        )

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.EntrypointDict], int]:
        """Fetch a list of entrypoints, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of entrypoints to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of entrypoints and the total number of entrypoints
            matching the query.

        Raises:
            SearchNotImplementedError: If a search string is provided.
            BackendDatabaseError: If the database query returns a None when counting
                the number of entrypoints.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of entrypoints")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.EntryPoint.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.EntryPoint.resource_snapshot_id,
            )
        )
        total_num_entrypoints = db.session.scalars(stmt).first()

        if total_num_entrypoints is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_entrypoints == 0:
            return [], total_num_entrypoints

        entrypoints_stmt = (
            select(models.EntryPoint)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.EntryPoint.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            entrypoints_stmt = entrypoints_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        entrypoints = list(db.session.scalars(entrypoints_stmt).unique().all())

        queue_ids = set(
            resource.resource_id
            for entrypoint in entrypoints
            for resource in entrypoint.children
            if resource.resource_type == "queue" and not resource.is_deleted
        )
        queues = {
            queue.resource_id: queue
            for queue in self._queue_ids_service.get(
                list(queue_ids), error_if_not_found=True
            )
        }

        entrypoint_dicts = {
            entrypoint.resource_id: utils.EntrypointDict(
                entry_point=entrypoint, queues=[], has_draft=False
            )
            for entrypoint in entrypoints
        }
        for entrypoint in entrypoint_dicts.values():
            for resource in entrypoint["entry_point"].children:
                if resource.resource_type == "queue" and not resource.is_deleted:
                    entrypoint["queues"].append(queues[resource.resource_id])

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(tuple(entrypoint_dicts.keys())),
            models.DraftResource.user_id == current_user.user_id,
        )
        for resource_id in db.session.scalars(drafts_stmt):
            entrypoint_dicts[resource_id]["has_draft"] = True

        return list(entrypoint_dicts.values()), total_num_entrypoints


class EntrypointIdService(object):
    """The service methods for creating and managing entrypoints by their unique id."""

    @inject
    def __init__(
        self,
        entrypoint_name_service: EntrypointNameService,
        queue_ids_service: QueueIdsService,
    ) -> None:
        """Initialize the entrypoint service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_name_service: A EntrypointNameService object.
            queue_ids_service: A QueueIdsService object.
        """
        self._entrypoint_name_service = entrypoint_name_service
        self._queue_ids_service = queue_ids_service

    def get(
        self,
        entrypoint_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.EntrypointDict | None:
        """Fetch a entrypoint by its unique id.

        Args:
            entrypoint_id: The unique id of the entrypoint.
            error_if_not_found: If True, raise an error if the entrypoint is not found.
                Defaults to False.

        Returns:
            The entrypoint object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the entrypoint is not found and
            `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get entrypoint by id", entrypoint_id=entrypoint_id)

        stmt = (
            select(models.EntryPoint)
            .join(models.Resource)
            .where(
                models.EntryPoint.resource_id == entrypoint_id,
                models.EntryPoint.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        entrypoint = db.session.scalars(stmt).first()

        if entrypoint is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    RESOURCE_TYPE, entrypoint_id=entrypoint_id
                )

            return None

        queue_ids = set(
            resource.resource_id
            for resource in entrypoint.children
            if resource.resource_type == "queue" and not resource.is_deleted
        )
        queues = self._queue_ids_service.get(list(queue_ids), error_if_not_found=True)

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == entrypoint.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.EntrypointDict(
            entry_point=entrypoint, queues=queues, has_draft=has_draft
        )

    def modify(
        self,
        entrypoint_id: int,
        name: str,
        description: str,
        task_graph: str,
        parameters: list[dict[str, Any]],
        queue_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.EntrypointDict | None:
        """Modify an entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.
            name: The new name of the entrypoint.
            description: The new description of the entrypoint.
            task_graph: The new task graph of the entrypoint.
            parameters: the new parameters of the entrypoint.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated entrypoint object.

        Raises:
            EntityDoesNotExistError: If the entrypoint is not found and
                `error_if_not_found` is True.
            EntityExistsError: If the entrypoint name already exists.
            QueryParameterNotUniqueError: If the values for the "name" parameter in the
                parameters list is not unique
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        duplicates = find_non_unique("name", parameters)
        if len(duplicates) > 0:
            raise QueryParameterNotUniqueError(RESOURCE_TYPE, name=duplicates)

        entrypoint_dict = self.get(
            entrypoint_id, error_if_not_found=error_if_not_found, log=log
        )

        if entrypoint_dict is None:
            return None

        entrypoint = entrypoint_dict["entry_point"]
        group_id = entrypoint.resource.group_id
        if name != entrypoint.name:
            duplicate = self._entrypoint_name_service.get(
                name, group_id=group_id, log=log
            )
            if duplicate is not None:
                raise EntityExistsError(
                    RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
                )

        queues = self._queue_ids_service.get(queue_ids, error_if_not_found=True)

        entrypoint_parameters = [
            models.EntryPointParameter(
                name=param["name"],
                default_value=param["default_value"],
                parameter_type=param["parameter_type"],
                parameter_number=i,
            )
            for i, param in enumerate(parameters)
        ]

        new_entrypoint = models.EntryPoint(
            name=name,
            description=description,
            task_graph=task_graph,
            parameters=entrypoint_parameters,
            resource=entrypoint.resource,
            creator=current_user,
        )
        new_entrypoint.entry_point_plugin_files = [
            models.EntryPointPluginFile(
                entry_point=new_entrypoint,
                plugin=entry_point_plugin_file.plugin,
                plugin_file=entry_point_plugin_file.plugin_file,
            )
            for entry_point_plugin_file in entrypoint.entry_point_plugin_files
        ]

        plugin_resources = list(
            {
                plugin.plugin.resource_id: plugin.plugin.resource
                for plugin in new_entrypoint.entry_point_plugin_files
            }.values()
        )
        queue_resources = [queue.resource for queue in queues]
        new_entrypoint.children = plugin_resources + queue_resources

        db.session.add(new_entrypoint)

        if commit:
            db.session.commit()
            log.debug(
                "Entrypoint modification successful",
                entrypoint_id=entrypoint_id,
                name=name,
                description=description,
            )

        return utils.EntrypointDict(
            entry_point=new_entrypoint, queues=queues, has_draft=False
        )

    def delete(self, entrypoint_id: int, **kwargs) -> dict[str, Any]:
        """Delete a entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=entrypoint_id, resource_type=RESOURCE_TYPE, is_deleted=False
        )
        entrypoint_resource = db.session.scalars(stmt).first()

        if entrypoint_resource is None:
            raise EntityDoesNotExistError(RESOURCE_TYPE, entrypoint_id=entrypoint_id)

        deleted_resource_lock = models.ResourceLock(
            resource_lock_type=resource_lock_types.DELETE,
            resource=entrypoint_resource,
        )
        db.session.add(deleted_resource_lock)
        db.session.commit()
        log.debug("Entrypoint deleted", entrypoint_id=entrypoint_id)

        return {"status": "Success", "id": [entrypoint_id]}


class EntrypointIdPluginsService(object):
    """The service methods for creating and managing entrypoints by their unique id."""

    @inject
    def __init__(
        self,
        entrypoint_id_service: EntrypointIdService,
        plugin_ids_service: PluginIdsService,
        queue_ids_service: QueueIdsService,
    ) -> None:
        """Initialize the entrypoint service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_id_service: A EntrypointIdService object.
            plugin_ids_service: A PluginIdsService object.
            queue_ids_service: A QueueIdsService object.
        """
        self._entrypoint_id_service = entrypoint_id_service
        self._plugin_ids_service = plugin_ids_service
        self._queue_ids_service = queue_ids_service

    def get(
        self,
        entrypoint_id: int,
        **kwargs,
    ) -> list[utils.PluginWithFilesDict]:
        """Fetch the plugin snapshots for an entrypoint by its unique id.

        Args:
            entrypoint_id: The unique id of the entrypoint.

        Returns:
            The plugin snapshots for the entrypoint.

        Raises:
            EntityDoesNotExistError: If the entrypoint is not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get entrypoint by id", entrypoint_id=entrypoint_id)

        entrypoint = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )

        return _get_entrypoint_plugin_snapshots(entrypoint["entry_point"])

    def append(
        self,
        entrypoint_id: int,
        plugin_ids: list[int],
        commit: bool = True,
        **kwargs,
    ) -> list[utils.PluginWithFilesDict]:
        """Append plugins to an entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.
            plugin_ids: The plugins to be appended. If a plugin is already attached
                to the entrypoint, it is synced to the latest snapshot.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated entrypoint object.

        Raises:
            EntityDoesNotExistError: If the entrypoint is not found.
            EntityExistsError: If the entrypoint name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        entrypoint = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )["entry_point"]

        new_entrypoint_parameters = [
            models.EntryPointParameter(
                name=param.name,
                default_value=param.default_value,
                parameter_type=param.parameter_type,
                parameter_number=param.parameter_number,
            )
            for param in entrypoint.parameters
        ]
        new_entrypoint = models.EntryPoint(
            name=entrypoint.name,
            description=entrypoint.description,
            task_graph=entrypoint.task_graph,
            parameters=new_entrypoint_parameters,
            resource=entrypoint.resource,
            creator=current_user,
        )

        new_entry_point_plugin_files = [
            models.EntryPointPluginFile(
                entry_point=new_entrypoint,
                plugin=plugin["plugin"],
                plugin_file=plugin_file,
            )
            for plugin in self._plugin_ids_service.get(
                plugin_ids, error_if_not_found=True
            )
            for plugin_file in plugin["plugin_files"]
        ]
        existing_entry_point_plugin_files = [
            models.EntryPointPluginFile(
                entry_point=new_entrypoint,
                plugin=entry_point_plugin_file.plugin,
                plugin_file=entry_point_plugin_file.plugin_file,
            )
            for entry_point_plugin_file in entrypoint.entry_point_plugin_files
            if entry_point_plugin_file.plugin.resource_id not in set(plugin_ids)
        ]
        new_entrypoint.entry_point_plugin_files = (
            existing_entry_point_plugin_files + new_entry_point_plugin_files
        )

        plugin_resources = list(
            {
                plugin.plugin.resource_id: plugin.plugin.resource
                for plugin in new_entrypoint.entry_point_plugin_files
            }.values()
        )
        queue_resources = [
            resource
            for resource in entrypoint.children
            if resource.resource_type == "queue"
        ]
        new_entrypoint.children = plugin_resources + queue_resources

        db.session.add(new_entrypoint)

        if commit:
            db.session.commit()
            log.debug(
                "Plugins appended to Entrypoint successfully",
                entrypoint_id=entrypoint_id,
                plugin_ids=plugin_ids,
            )

        return _get_entrypoint_plugin_snapshots(new_entrypoint)


class EntrypointIdPluginsIdService(object):
    """The service methods for creating and managing entrypoints by their unique id."""

    @inject
    def __init__(
        self,
        entrypoint_id_service: EntrypointIdService,
        queue_ids_service: QueueIdsService,
    ) -> None:
        """Initialize the entrypoint service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_id_service: A EntrypointIdService object.
            queue_ids_service: A QueueIdsService object.
        """
        self._entrypoint_id_service = entrypoint_id_service
        self._queue_ids_service = queue_ids_service

    def get(
        self,
        entrypoint_id: int,
        plugin_id: int,
        **kwargs,
    ) -> utils.PluginWithFilesDict:
        """Fetch the plugin snapshots for an entrypoint by its unique id.

        Args:
            entrypoint_id: The unique id of the entrypoint.

        Returns:
            The plugin snapshots for the entrypoint.

        Raises:
            EntityDoesNotExistError: If the entrypoint or plugin is not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get entrypoint by id", entrypoint_id=entrypoint_id)

        entrypoint = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )["entry_point"]

        plugins = {
            entry_point_plugin_file.plugin.resource_id: entry_point_plugin_file.plugin
            for entry_point_plugin_file in entrypoint.entry_point_plugin_files
        }
        if plugin_id not in plugins:
            raise EntityDoesNotExistError(
                PLUGIN_RESOURCE_TYPE, entrypoint_id=entrypoint_id, plugin_id=plugin_id
            )

        plugin = utils.PluginWithFilesDict(
            plugin=plugins[plugin_id], plugin_files=[], has_draft=None
        )
        for entry_point_plugin_file in entrypoint.entry_point_plugin_files:
            if entry_point_plugin_file.plugin.resource_id == plugin_id:
                plugin["plugin_files"].append(entry_point_plugin_file.plugin_file)

        return plugin

    def delete(
        self,
        entrypoint_id: int,
        plugin_id: int,
        commit: bool = True,
        **kwargs,
    ) -> dict[str, Any]:
        """Remove a plugin from an entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.
            plugin_id: The plugin to be removed.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            A dictionary reporting the status of the request.

        Raises:
            EntityDoesNotExistError: If the entrypoint or plugin is not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        entrypoint = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )["entry_point"]

        plugin_ids = set(
            plugin.plugin.resource_id for plugin in entrypoint.entry_point_plugin_files
        )
        if plugin_id not in plugin_ids:
            raise EntityDoesNotExistError(
                PLUGIN_RESOURCE_TYPE, entrypoint_id=entrypoint_id, plugin_id=plugin_id
            )

        # create a new snapshot with the plugin removed
        entrypoint_parameters = [
            models.EntryPointParameter(
                name=param.name,
                default_value=param.default_value,
                parameter_type=param.parameter_type,
                parameter_number=param.parameter_number,
            )
            for param in entrypoint.parameters
        ]
        new_entrypoint = models.EntryPoint(
            name=entrypoint.name,
            description=entrypoint.description,
            task_graph=entrypoint.task_graph,
            parameters=entrypoint_parameters,
            resource=entrypoint.resource,
            creator=current_user,
        )
        new_entrypoint.entry_point_plugin_files = [
            models.EntryPointPluginFile(
                entry_point=new_entrypoint,
                plugin=entry_point_plugin_file.plugin,
                plugin_file=entry_point_plugin_file.plugin_file,
            )
            for entry_point_plugin_file in entrypoint.entry_point_plugin_files
            if entry_point_plugin_file.plugin.resource_id != plugin_id
        ]

        # remove the plugin resource dependency association
        new_entrypoint.children = [
            resource
            for resource in entrypoint.children
            if resource.resource_id != plugin_id
        ]

        db.session.add(new_entrypoint)

        if commit:
            db.session.commit()
            log.debug(
                "Plugin removed from entrypoint",
                entrypoint_id=entrypoint_id,
                plugin_id=plugin_id,
            )

        return {"status": "Success", "id": [plugin_id]}


class EntrypointIdsService(object):
    """The service methods for retrieving entrypoints from a list of ids."""

    def get(
        self,
        entrypoint_ids: list[int],
        error_if_not_found: bool = False,
        **kwargs,
    ) -> list[models.EntryPoint]:
        """Fetch a list of entrypoints by their unique ids.

        Args:
            entrypoint_ids: The unique ids of the entrypoints.
            error_if_not_found: If True, raise an error if the entrypoint is not found.
                Defaults to False.

        Returns:
            The entrypoint object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the entrypoint is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get entrypoint by id", entrypoint_ids=entrypoint_ids)

        stmt = (
            select(models.EntryPoint)
            .join(models.Resource)
            .where(
                models.EntryPoint.resource_id.in_(tuple(entrypoint_ids)),
                models.EntryPoint.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        entrypoints = list(db.session.scalars(stmt).all())

        if len(entrypoints) != len(entrypoint_ids) and error_if_not_found:
            entrypoint_ids_missing = set(entrypoint_ids) - set(
                entrypoint.resource_id for entrypoint in entrypoints
            )
            raise EntityDoesNotExistError(
                RESOURCE_TYPE, entrypoint_ids=list(entrypoint_ids_missing)
            )

        return entrypoints


class EntrypointIdQueuesService(object):
    """The service methods for managing queues attached to an entrypoint."""

    @inject
    def __init__(
        self,
        entrypoint_id_service: EntrypointIdService,
        queue_ids_service: QueueIdsService,
    ) -> None:
        """Initialize the entrypoint service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_id_service: A EntrypointIdService object.
            queue_ids_service: A QueueIdsService object.
        """
        self._entrypoint_id_service = entrypoint_id_service
        self._queue_ids_service = queue_ids_service

    def get(
        self,
        entrypoint_id: int,
        **kwargs,
    ) -> list[models.Queue]:
        """Fetch the list of queues for an entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.
            error_if_not_found: If True, raise an error if the entrypoint is not found.
                Defaults to False.

        Returns:
            The list of plugins.

        Raises:
            EntityDoesNotExistError: If the entrypoint is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Get queues for an entrypoint by resource id", resource_id=entrypoint_id
        )

        entrypoint_dict = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )
        return entrypoint_dict["queues"]

    def append(
        self,
        entrypoint_id: int,
        queue_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> list[models.Queue] | None:
        """Append one or more Queues to an entrypoint

        Args:
            entrypoint_id: The unique id of the entrypoint.
            queue_ids: The list of queue ids to append.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated list of queues resource objects.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
            EntityDoesNotExistError: If one or more queues are not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Append queues to an entrypoint by resource id", resource_id=entrypoint_id
        )

        entrypoint_dict = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )
        entrypoint = entrypoint_dict["entry_point"]
        queues = entrypoint_dict["queues"]

        existing_queue_ids = set(
            resource.resource_id
            for resource in entrypoint.children
            if resource.resource_type == "queue"
        )
        new_queue_ids = set(queue_ids) - existing_queue_ids
        new_queues = self._queue_ids_service.get(
            list(new_queue_ids), error_if_not_found=True, log=log
        )

        entrypoint.children.extend([queue.resource for queue in new_queues])

        if commit:
            db.session.commit()
            log.debug("Queues appended successfully", queue_ids=queue_ids)

        return queues + new_queues

    def modify(
        self,
        entrypoint_id: int,
        queue_ids: list[int],
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> list[models.Queue]:
        """Modify the list of queues for an entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.
            queue_ids: The list of queue ids to append.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated queue resource object.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
            EntityDoesNotExistError: If one or more queues are not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        entrypoint_dict = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )
        entrypoint = entrypoint_dict["entry_point"]

        queues = self._queue_ids_service.get(
            queue_ids, error_if_not_found=True, log=log
        )

        plugin_resources = [
            resource
            for resource in entrypoint.children
            if resource.resource_type == "plugin"
        ]
        queue_resources = [queue.resource for queue in queues]
        entrypoint.children = plugin_resources + queue_resources

        if commit:
            db.session.commit()
            log.debug("Entrypoint queues updated successfully", queue_ids=queue_ids)

        return queues

    def delete(self, entrypoint_id: int, **kwargs) -> dict[str, Any]:
        """Remove queues from an entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        entrypoint_dict = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )
        entrypoint = entrypoint_dict["entry_point"]
        entrypoint.children = [
            resource
            for resource in entrypoint.children
            if resource.resource_type == "plugin"
        ]

        queue_ids = [queue.resource_id for queue in entrypoint_dict["queues"]]

        db.session.commit()
        log.debug(
            "Queues removed from entrypoint",
            entrypoint_id=entrypoint_id,
            queue_ids=queue_ids,
        )

        return {"status": "Success", "id": queue_ids}


class EntrypointIdQueuesIdService(object):
    """The service methods for removing a queue attached to an entrypoint."""

    @inject
    def __init__(
        self,
        entrypoint_id_service: EntrypointIdService,
    ) -> None:
        """Initialize the entrypoint service.

        All arguments are provided via dependency injection.

        Args:
            entrypoint_id_service: A EntrypointIdService object.
        """
        self._entrypoint_id_service = entrypoint_id_service

    def delete(self, entrypoint_id: int, queue_id, **kwargs) -> dict[str, Any]:
        """Remove a queue from an entrypoint.

        Args:
            entrypoint_id: The unique id of the entrypoint.
            queue_id: The unique id of the queue.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        entrypoint_dict = cast(
            utils.EntrypointDict,
            self._entrypoint_id_service.get(
                entrypoint_id, error_if_not_found=True, log=log
            ),
        )
        entrypoint = entrypoint_dict["entry_point"]

        queue_resources = {
            resource.resource_id: resource
            for resource in entrypoint.children
            if resource.resource_type == "queue"
        }

        removed_queue = queue_resources.pop(queue_id, None)

        if removed_queue is None:
            raise EntityDoesNotExistError(QUEUE_RESOURCE_TYPE, queue_id=queue_id)

        plugin_resources = [
            resource
            for resource in entrypoint.children
            if resource.resource_type == "plugin"
        ]
        entrypoint.children = plugin_resources + list(queue_resources.values())

        db.session.commit()
        log.debug("Queue removed", entrypoint_id=entrypoint_id, queue_id=queue_id)

        return {"status": "Success", "id": [queue_id]}


class EntrypointNameService(object):
    """The service methods for managing entrypoints by their name."""

    def get(
        self,
        name: str,
        group_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.EntryPoint | None:
        """Fetch a entrypoint by its name.

        Args:
            name: The name of the entrypoint.
            group_id: The the group id of the entrypoint.
            error_if_not_found: If True, raise an error if the entrypoint is not found.
                Defaults to False.

        Returns:
            The entrypoint object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the entrypoint is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get entrypoint by name", entrypoint_name=name, group_id=group_id)

        stmt = (
            select(models.EntryPoint)
            .join(models.Resource)
            .where(
                models.EntryPoint.name == name,
                models.Resource.group_id == group_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.EntryPoint.resource_snapshot_id,
            )
        )
        entrypoint = db.session.scalars(stmt).first()

        if entrypoint is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    RESOURCE_TYPE, name=name, group_id=group_id
                )

            return None

        return entrypoint


def _get_entrypoint_plugin_snapshots(
    entrypoint: models.EntryPoint,
) -> list[utils.PluginWithFilesDict]:
    plugins_dict = {
        entry_point_plugin_file.plugin.resource_id: utils.PluginWithFilesDict(
            plugin=entry_point_plugin_file.plugin, plugin_files=[], has_draft=False
        )
        for entry_point_plugin_file in entrypoint.entry_point_plugin_files
    }
    for entry_point_plugin_file in entrypoint.entry_point_plugin_files:
        resource_id = entry_point_plugin_file.plugin.resource_id
        plugin_file = entry_point_plugin_file.plugin_file
        plugins_dict[resource_id]["plugin_files"].append(plugin_file)
    return list(plugins_dict.values())



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import collections
import itertools
import logging
from collections.abc import Iterable, Mapping, MutableMapping, Sequence
from typing import Any, Union

import dioptra.pyplugs
from dioptra.sdk.exceptions.task_engine import (
    IllegalOutputReferenceError,
    IllegalPluginNameError,
    MissingGlobalParametersError,
    MissingTaskPluginNameError,
    NonIterableTaskOutputError,
    OutputNotFoundError,
    StepError,
    StepNotFoundError,
    TaskPluginNotFoundError,
    UnresolvableReferenceError,
)
from dioptra.task_engine import util


def _get_logger() -> logging.Logger:
    """
    Get a logger to use for functions in this module.

    Returns:
        The logger
    """
    return logging.getLogger(__name__)


def _resolve_reference(
    reference: str,
    global_parameters: Mapping[str, Any],
    step_outputs: Mapping[str, Mapping[str, Any]],
) -> Any:
    """
    Resolve a reference to a task output or global parameter.

    Args:
        reference: The reference to resolve, without the "$" prefix
        global_parameters: The global parameters in use for this run, as a
            mapping from parameter name to value
        step_outputs: The step outputs we have thus far.  This is a nested
            mapping: step name => output name => output value.

    Returns:
        The referenced value
    """

    if "." in reference:
        # Must be an <step>.<output> formatted reference
        step_name, output_name = reference.split(".", 1)

        step_output = step_outputs.get(step_name)
        if not step_output:
            raise StepNotFoundError(step_name)

        if output_name not in step_output:
            raise OutputNotFoundError(step_name, output_name)

        value = step_output[output_name]

    else:
        # A bare name may refer to either a global parameter or the only
        # output of a step.  Let's assume prior validation ensured the same
        # name does not occur in both places.
        if reference in global_parameters:
            value = global_parameters[reference]
        elif reference in step_outputs:
            outputs = step_outputs[reference]
            if len(outputs) != 1:
                raise IllegalOutputReferenceError(reference)
            value = next(iter(outputs.values()))
        else:
            raise UnresolvableReferenceError(reference)

    return value


def _resolve_task_parameter_value(
    arg_spec: Any,
    global_parameters: Mapping[str, Any],
    step_outputs: Mapping[str, Mapping[str, Any]],
) -> Any:
    """
    Resolve a specification for one argument of a task invocation, to the
    actual value to be used in the invocation.

    Args:
        arg_spec: The argument specification
        global_parameters: The global parameters in use for this run, as a
            mapping from parameter name to value
        step_outputs: The step outputs we have thus far.  This is a nested
            mapping: step name => output name => output value.

    Returns:
        The value to use for the given parameter
    """

    if isinstance(arg_spec, str):
        if util.is_reference(arg_spec):
            arg_value = _resolve_reference(
                arg_spec[1:], global_parameters, step_outputs
            )

        elif arg_spec.startswith("$$"):
            # "escaped" dollar sign: replace only the initial "$$" with "$"
            arg_value = arg_spec[1:]

        else:
            arg_value = arg_spec

    elif isinstance(arg_spec, dict):
        arg_value = {
            key: _resolve_task_parameter_value(value, global_parameters, step_outputs)
            for key, value in arg_spec.items()
        }

    elif isinstance(arg_spec, list):
        arg_value = [
            _resolve_task_parameter_value(sub_val, global_parameters, step_outputs)
            for sub_val in arg_spec
        ]

    else:
        arg_value = arg_spec

    return arg_value


def _positional_specs_to_args(
    arg_specs: Iterable[Any],
    global_parameters: Mapping[str, Any],
    step_outputs: Mapping[str, Mapping[str, Any]],
) -> list[Any]:
    """
    Resolve a positional parameter style invocation specification to a list
    of actual parameter values to use in the task invocation.

    Args:
        arg_specs: A list of argument specifications.
        global_parameters: The global parameters in use for this run, as a
            mapping from parameter name to value
        step_outputs: The step outputs we have thus far.  This is a nested
            mapping: step name => output name => output value.

    Returns:
        A list of values to use for the invocation
    """

    arg_values = [
        _resolve_task_parameter_value(arg_spec, global_parameters, step_outputs)
        for arg_spec in arg_specs
    ]

    return arg_values


def _kwarg_specs_to_kwargs(
    kwarg_specs: Mapping[str, Any],
    global_parameters: Mapping[str, Any],
    step_outputs: Mapping[str, Mapping[str, Any]],
) -> dict[str, Any]:
    """
    Resolve a keyword arg style invocation specification to a mapping of actual
    parameter values to use in the task invocation.

    Args:
        kwarg_specs: A mapping from keyword arg name to value specification
        global_parameters: The global parameters in use for this run, as a
            mapping from parameter name to value
        step_outputs: The step outputs we have thus far.  This is a nested
            mapping: step name => output name => output value.

    Returns:
        A mapping from parameter name to actual value to use for the
        task invocation
    """
    kwarg_values = {
        kwarg_name: _resolve_task_parameter_value(
            kwarg_value, global_parameters, step_outputs
        )
        for kwarg_name, kwarg_value in kwarg_specs.items()
    }

    return kwarg_values


def _get_invocation_args(
    step: Mapping[str, Any],
    global_parameters: Mapping[str, Any],
    step_outputs: Mapping[str, Mapping[str, Any]],
) -> tuple[list[Any], dict[str, Any]]:
    """
    Resolve a task invocation specification to all of the positional and
    keyword arg values to use in the invocation.

    Args:
        step: The step description
        global_parameters: The global parameters in use for this run, as a
            mapping from parameter name to value
        step_outputs: The step outputs we have thus far.  This is a nested
            mapping: step name => output name => output value.

    Returns:
        A 2-tuple including a list of positional values to use in the
        task invocation, followed by a mapping with keyword arg names and
        values.
    """

    pos_arg_specs, kwarg_specs = util.step_get_invocation_arg_specs(step)

    # step_get_invocation_args() is written to be graceful in the face of a
    # malformed step definition (and return nulls), but at this point I think
    # we can assume validation has already been done, as would be the normal
    # workflow, so nulls won't happen here.
    assert pos_arg_specs is not None
    assert kwarg_specs is not None

    # Assume for now that validation has completed successfully, so we always
    # have a correct step definition with arg specs?
    arg_values = _positional_specs_to_args(
        pos_arg_specs, global_parameters, step_outputs
    )

    kwarg_values = _kwarg_specs_to_kwargs(kwarg_specs, global_parameters, step_outputs)

    return arg_values, kwarg_values


def _update_output_map(
    step_outputs: MutableMapping[str, MutableMapping[str, Any]],
    step_name: str,
    output_defs: Union[Mapping[str, str], Sequence[Mapping[str, str]]],
    new_outputs: Any,
) -> None:
    """
    Update the step outputs mapping according to task metadata regarding output
    name(s), and actual task plugin return value(s).

    Args:
        step_outputs: The step outputs we have thus far.  This a is nested
            mapping: step name => output name => output value.
        step_name: The name of the step for which we have output values
        output_defs: Task metadata regarding outputs.  This may be either a
            single or list of mappings.
        new_outputs: The output(s) from the completed step
    """

    log = _get_logger()

    if isinstance(output_defs, Mapping):
        # If a single output is defined, store the task return value directly
        # under that name.
        output_name = next(iter(output_defs))
        step_outputs[step_name][output_name] = new_outputs

    else:
        # else, output_defs is a list.  Task plugin return value must
        # be iterable.
        if not util.is_iterable(new_outputs):
            raise NonIterableTaskOutputError(new_outputs, step_name)

        # Support more general iterables as return values from tasks, which may
        # not be len()-able.  If we can get a length, then we can sanity check
        # the number of output names given against the number of output values
        # produced by the task, and produce a warning if they don't match.
        try:
            num_outputs = len(new_outputs)
        except TypeError:
            num_outputs = None

        num_output_names = len(output_defs)

        if num_outputs is not None and num_outputs != num_output_names:
            log.warning(
                'Different numbers of outputs and output definitions for step "%s": %d != %d',  # noqa: B950
                step_name,
                num_outputs,
                num_output_names,
            )

        output_names = itertools.chain.from_iterable(output_defs)

        for output_name, output_value in zip(output_names, new_outputs):
            step_outputs[step_name][output_name] = output_value


def _resolve_global_parameters(
    global_parameter_spec: Mapping[str, Any],
    global_parameters: MutableMapping[str, Any],
) -> None:
    """
    Build a complete set of global parameters from the specification given in
    the experiment description and the parameter names and values given from an
    external source.

    This will update global_parameters rather than creating and returning a new
    mapping.  Entries are added for params which were missing and for which
    defaults were given in the description.  Any params given which were not
    defined in the description are removed from the mapping.  Extraneous
    parameters are not considered an error.  Any parameters defined in the
    description for which a value can't be obtained, will produce an error.

    Args:
        global_parameter_spec: The global parameter spec from the experiment
            description
        global_parameters: A mapping from parameter name to value, representing
            global parameters given from some external source, e.g. on a
            commandline.
    """

    log = _get_logger()

    missing_parameters = []

    for param_name, param_def in global_parameter_spec.items():
        if param_name not in global_parameters:
            if isinstance(param_def, dict):
                if "default" in param_def:
                    global_parameters[param_name] = param_def["default"]
                else:
                    missing_parameters.append(param_name)
            else:
                global_parameters[param_name] = param_def

    extra_parameters = global_parameters.keys() - global_parameter_spec.keys()

    if missing_parameters:
        raise MissingGlobalParametersError(missing_parameters)

    if extra_parameters:
        # This doesn't need to be a showstopper error I think.
        log.warning(
            "Some global parameters were unused: %s", ", ".join(extra_parameters)
        )

        # Maybe should also remove extras from the mapping?
        for param_name in extra_parameters:
            del global_parameters[param_name]


def _get_pyplugs_coords(task_plugin: str) -> list[str]:
    """
    Split a fully qualified task plugin to the three parts required as
    identifying coordinates by pyplugs.  The coordinates are:

        <package> <module> <function name>

    The task plugin is a dot-delimited string with at least two components.
    The last two components map to <module> and <function name>; everything to
    their left comprises the <package> part.  If there are only two components
    in the plugin, the package will default to the empty string.  This keeps
    with how pyplugs registers plugins.  For example:

        a.b => "" a b
        a.b.c => a b c
        a.b.c.d => a.b c d
        a.b.c.d.e => a.b.c d e

        etc...

    Args:
        task_plugin: The dotted plugin string from the declarative experiment
            description

    Returns:
        A length-3 list of pyplugs coordinates
    """
    coords = task_plugin.rsplit(".", 2)

    if len(coords) < 2:
        raise IllegalPluginNameError(task_plugin)

    elif len(coords) == 2:
        coords.insert(0, "")

    return coords


def _run_step(
    step: Mapping[str, Any],
    task_plugin_id: str,
    global_parameters: Mapping[str, Any],
    step_outputs: Mapping[str, Mapping[str, Any]],
) -> Any:
    """
    Run one step of a task graph.

    Args:
        step: The step description
        task_plugin_id: The task plugin to call, in a composed dotted
            string form with all the parts needed by pyplugs, e.g. "a.b.c.d"
        global_parameters: The global parameters in use for this run, as a
            mapping from parameter name to value
        step_outputs: The step outputs we have thus far.  This is a nested
            mapping: step name => output name => output value.

    Returns:
        The step output (i.e. whatever the task plugin returned)
    """

    log = _get_logger()

    arg_values, kwarg_values = _get_invocation_args(
        step, global_parameters, step_outputs
    )

    if arg_values:
        log.debug("args: %s", arg_values)
    if kwarg_values:
        log.debug("kwargs: %s", kwarg_values)

    package_name, module_name, func_name = _get_pyplugs_coords(task_plugin_id)

    output = dioptra.pyplugs.call(
        package_name, module_name, func_name, *arg_values, **kwarg_values
    )

    return output


def run_experiment(
    experiment_desc: Mapping[str, Any], global_parameters: MutableMapping[str, Any]
) -> None:
    """
    Run an experiment via a declarative experiment description.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent
        global_parameters: External parameter values to use in the
            experiment, as a dict
    """

    log = _get_logger()

    global_parameter_spec = experiment_desc.get("parameters", {})
    tasks = experiment_desc["tasks"]
    graph = experiment_desc["graph"]

    _resolve_global_parameters(global_parameter_spec, global_parameters)

    if log.isEnabledFor(logging.DEBUG):
        props_values = "\n  ".join(
            param_name + ": " + str(param_value)
            for param_name, param_value in global_parameters.items()
        )
        log.debug("Global parameters:\n  %s", props_values)

    step_outputs: MutableMapping[str, MutableMapping[str, Any]] = (
        collections.defaultdict(dict)
    )

    step_order = util.get_sorted_steps(graph)

    log.debug("Step order:\n  %s", "\n  ".join(step_order))

    for step_name in step_order:
        try:
            log.info("Running step: %s", step_name)

            step = graph[step_name]

            task_plugin_short_name = util.step_get_plugin_short_name(step)
            if not task_plugin_short_name:
                raise MissingTaskPluginNameError(step_name)

            task_def = tasks.get(task_plugin_short_name)
            if not task_def:
                raise TaskPluginNotFoundError(task_plugin_short_name, step_name)

            task_plugin_id = task_def["plugin"]

            output = _run_step(step, task_plugin_id, global_parameters, step_outputs)

            output_defs = task_def.get("outputs")
            if output_defs:
                _update_output_map(step_outputs, step_name, output_defs, output)
                log.debug("Output(s): %s", str(step_outputs[step_name]))

            # else: should I warn if there was an output from the task but no
            # output_names were given?

        except StepError as e:
            # Fill in useful contextual info on the error if necessary.
            if not e.context_step_name:
                e.context_step_name = step_name
            raise



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import json
from typing import Any, Final, cast

JsonType = dict[str, Any] | list[Any]
GlobalParameterType = str | float | int | bool | dict[str, Any] | list[Any] | None


BOOLEAN_PARAM_TYPE: Final[str] = "boolean"
FLOAT_PARAM_TYPE: Final[str] = "float"
INTEGER_PARAM_TYPE: Final[str] = "integer"
LIST_PARAM_TYPE: Final[str] = "list"
MAPPING_PARAM_TYPE: Final[str] = "mapping"
STRING_PARAM_TYPE: Final[str] = "string"


def coerce_to_type(x: str | None, type_name: str) -> GlobalParameterType:
    coerce_fn_registry = {
        BOOLEAN_PARAM_TYPE: to_boolean_type,
        FLOAT_PARAM_TYPE: to_float_type,
        INTEGER_PARAM_TYPE: to_integer_type,
        LIST_PARAM_TYPE: to_list_type,
        MAPPING_PARAM_TYPE: to_mapping_type,
        STRING_PARAM_TYPE: to_string_type,
    }

    if type_name not in coerce_fn_registry:
        raise ValueError(f"Invalid parameter type: {type_name}.")

    if x is None:
        return None

    coerce_fn = coerce_fn_registry[type_name]
    return cast(GlobalParameterType, coerce_fn(x))


def to_string_type(x: str) -> str:
    return x


def to_boolean_type(x: str) -> bool:
    if x.lower() not in {"true", "false"}:
        raise ValueError(f"Not a boolean: {x}")

    return x.lower() == "true"


def to_float_type(x: str) -> float:
    # TODO: Handle coercion failures
    return float(x)


def to_integer_type(x: str) -> int:
    # TODO: Handle coercion failures
    return int(x)


def to_list_type(x: str) -> list[Any]:
    # TODO: Handle coercion failures
    x_coerced = cast(JsonType, json.loads(x))

    if not isinstance(x_coerced, list):
        raise ValueError(f"Not a list: {x}")

    return x_coerced


def to_mapping_type(x: str) -> dict[str, Any]:
    # TODO: Handle coercion failures
    x_coerced = cast(JsonType, json.loads(x))

    if not isinstance(x_coerced, dict):
        raise ValueError(f"Not a mapping: {x}")

    return x_coerced
